from opencopilot.eval.entities import (
    RetrievalExample,
    RetrievalResult,
    RetrievalSingleEvaluation,
    RetrievalSummaryEvaluation,
    RetrievalDataset,
)
from typing import List


def get_confusion_matrix(
    ground_truth: RetrievalExample, predicted: RetrievalResult
) -> RetrievalSingleEvaluation:
    """Compare a single ground-truth retrieval example with a single predicted result."""
    tp, fp, _, fn = [], [], [], []  # we won't calculate true-negatives

    for predicted_doc in predicted.documents:
        if predicted_doc in ground_truth.documents:
            tp.append(predicted_doc)
        else:
            fp.append(predicted_doc)

    for relevant_doc in ground_truth.documents:
        if relevant_doc not in predicted.documents:
            fn.append(relevant_doc)

    return tp, fp, fn


def get_summary_evaluation(
    single_evaluations: List[RetrievalSingleEvaluation],
) -> RetrievalSummaryEvaluation:
    """Roll up a list of single evaluations into a summary evaluation."""
    precisions = [e.precision for e in single_evaluations]
    recalls = [e.recall for e in single_evaluations]

    return RetrievalSummaryEvaluation(
        single_evaluations=single_evaluations,
        average_precision=sum(precisions) / len(precisions),
        average_recall=sum(recalls) / len(recalls),
    )


def evaluate_retrieval_single(
    ground_truth: RetrievalExample, predicted: RetrievalResult
) -> RetrievalSingleEvaluation:
    tp, fp, fn = get_confusion_matrix(ground_truth, predicted)
    n_tp, n_fp, n_fn = len(tp), len(fp), len(fn)

    if (n_tp + n_fp) == 0:
        print(f"Warning: No results from Weaviate for question: {ground_truth.query}")
        precision = 1.0
    else:
        precision = n_tp / (n_tp + n_fp)

    if (n_tp + n_fn) == 0:
        print(f"Warning: This example has zero expected URLs: {ground_truth.query}")
        recall = 1.0
    else:
        recall = n_tp / (n_tp + n_fn)

    return RetrievalSingleEvaluation(
        example=ground_truth,
        result=predicted,
        true_positives=tp,
        false_positives=fp,
        false_negatives=fn,
        precision=precision,
        recall=recall,
    )


def evaluate_retrieval_dataset(
    dataset: RetrievalDataset, predictions: List[RetrievalResult]
) -> RetrievalSummaryEvaluation:
    assert len(dataset.examples) == len(predictions)

    evals = []
    for i, prediction in enumerate(predictions):
        evals.append(evaluate_retrieval_single(dataset.examples[i], prediction))

    return get_summary_evaluation(evals)



import json
import os
from typing import Dict
from typing import List
from uuid import UUID

from langchain.schema import Document

from opencopilot import settings
from opencopilot.logger import api_logger

logger = api_logger.get()


class ConversationLogsRepositoryLocal:
    def __init__(self, conversation_logs_dir: str = ""):
        if not conversation_logs_dir:
            conversation_logs_dir = os.path.join(
                settings.get().LOGS_DIR, "conversation_logs"
            )
        self.conversation_logs_dir = conversation_logs_dir

    def log_prompt_template(
        self,
        conversation_id: UUID,
        message: str,
        prompt_template: str,
        response_message_id: str,
        token_count: int = None,
    ) -> None:
        self._append_to_file(
            conversation_id,
            {
                "response_message_id": response_message_id,
                "message": message,
                "prompt_template": prompt_template,
                "token_count": token_count,
            },
        )

    def log_prompt_text(
        self,
        conversation_id: UUID,
        message: str,
        prompt_text: str,
        response_message_id: str,
        token_count: int = None,
    ) -> None:
        self._append_to_file(
            conversation_id,
            {
                "response_message_id": response_message_id,
                "message": message,
                "prompt_text": prompt_text,
                "token_count": token_count,
            },
        )

    def log_history(
        self,
        conversation_id: UUID,
        message: str,
        history: str,
        response_message_id: str,
        token_count: int = None,
    ) -> None:
        self._append_to_file(
            conversation_id,
            {
                "response_message_id": response_message_id,
                "message": message,
                "history": history,
                "token_count": token_count,
            },
        )

    def log_context(
        self,
        conversation_id: UUID,
        message: str,
        contexts: List[Document],
        response_message_id: str,
        token_count: int = None,
    ) -> None:
        self._append_to_file(
            conversation_id,
            {
                "response_message_id": response_message_id,
                "message": message,
                "context": json.dumps([c.dict() for c in contexts]),
                "token_count": token_count,
            },
        )

    def get_logs_by_message(
        self,
        conversation_id: UUID,
        response_message_id: str,
    ) -> List[Dict]:
        try:
            with open(self._get_file_path(conversation_id), "r") as file:
                lines = file.readlines()
            formatted_lines: List[Dict] = []
            for line in lines:
                if not isinstance(line, str):
                    continue
                formatted_line = json.loads(line)
                if formatted_line.get("response_message_id") == response_message_id:
                    formatted_lines.append(formatted_line)
            return formatted_lines
        except:
            return []

    def remove_conversation(self, conversation_id: UUID):
        file_path = self._get_file_path(conversation_id)
        if os.path.exists(file_path):
            try:
                os.remove(file_path)
            except:
                pass

    def _get_file_path(self, conversation_id: UUID):
        return os.path.join(self.conversation_logs_dir, str(conversation_id)) + ".jsonl"

    def _append_to_file(self, conversation_id: UUID, log: Dict) -> None:
        try:
            if not os.path.exists(self.conversation_logs_dir):
                os.makedirs(self.conversation_logs_dir, exist_ok=True)
            with open(self._get_file_path(conversation_id), "a") as file:
                file.write(json.dumps(log) + "\n")
        except:
            logger.error(
                f"Error log_prompt_text, conversation_id: {str(conversation_id)}"
            )



import csv
import json
import os
import re
from typing import List

from langchain.document_loaders import CSVLoader
from langchain.document_loaders import PyPDFLoader
from langchain.document_loaders import TextLoader
from langchain.document_loaders import UnstructuredExcelLoader
from langchain.document_loaders import UnstructuredFileLoader
from langchain.schema import Document
from langchain.text_splitter import TextSplitter

from opencopilot import settings
from opencopilot.repository.documents import split_documents_use_case


def execute(
    data_dir: str, is_loading_deprecated: bool, text_splitter: TextSplitter
) -> List[Document]:
    files = []
    if not data_dir or not os.path.isdir(data_dir):
        return []
    ignore_files = [".DS_Store"]
    for dir_path, dir_names, file_names in os.walk(data_dir):
        for file_name in file_names:
            if file_name not in ignore_files:
                files.append(os.path.join(dir_path, file_name))

    files_count = len(files)
    if files_count > 0:
        print(f"[Loading {files_count} file{'s'[:files_count ^ 1]} from {data_dir}.]")
    else:
        raise Exception(f"[Add files to {data_dir}.]")

    documents = []
    for file_path in files:
        new_documents = []
        if (
            file_size := _get_file_size(file_path)
        ) > settings.get().MAX_DOCUMENT_SIZE_MB:
            print(
                f"Document {file_path} too big ({file_size} > {settings.get().MAX_DOCUMENT_SIZE_MB}), skipping."
            )
            continue
        if file_path.endswith(".csv"):
            with open(file_path, newline="") as csvfile:
                dialect = csv.Sniffer().sniff(csvfile.read(1024))

            loader = CSVLoader(
                file_path=file_path,
                csv_args={
                    "delimiter": dialect.delimiter,
                },
            )
            new_documents = loader.load()
        elif file_path.endswith(".tsv"):
            loader = CSVLoader(
                file_path=file_path,
                csv_args={
                    "delimiter": "\t",
                },
            )
            new_documents = loader.load()
        elif file_path.endswith(".pdf"):
            loader = PyPDFLoader(file_path)
            new_documents = loader.load()
        elif file_path.endswith(".xls") or file_path.endswith(".xlsx"):
            loader = UnstructuredExcelLoader(file_path)
            new_documents = loader.load()
        elif file_path.endswith(".json") and os.path.basename(file_path).startswith(
            "serialized_documents_"
        ):
            with open(file_path, "r") as f:
                document_dicts = json.load(f)
            for document in document_dicts:
                metadata = document["metadata"]
                if "deprecated" in metadata.get("source", ""):
                    if not is_loading_deprecated:
                        # skip deprecated information
                        continue
                    document["metadata"]["deprecated"] = True
                content = _cleanup_document(document["page_content"])
                new_documents.append(Document(page_content=content, metadata=metadata))
        elif file_path.endswith(".json"):
            loader = TextLoader(file_path)
            new_documents = loader.load()
        else:
            try:
                loader = UnstructuredFileLoader(file_path)
                new_documents = loader.load()
            except Exception as e:
                print(f"Error loading {file_path}")
        if text_splitter:
            document_chunks = split_documents_use_case.execute(
                text_splitter, new_documents
            )
            print(
                f"Generated {len(document_chunks)} document chunks from {len(new_documents)} documents"
            )
            documents.extend(document_chunks)
        else:
            print(f"Generated {len(new_documents)} documents")
            documents.extend(new_documents)
    return documents


def _cleanup_document(text: str) -> str:
    # cleanup special unicode characters + trailing newline
    pattern = r"[\u00A0-\uFFFF]\n?"
    clean = re.sub(pattern, "", text)
    clean = re.sub("\n+", "\n", clean)
    return clean


def _get_file_size(file_path):
    size_in_bytes = os.path.getsize(file_path)
    size_in_mb = size_in_bytes / 1024 / 1024
    return size_in_mb



from starlette import status


class APIErrorResponse(Exception):
    """Base class for other exceptions"""

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        raise NotImplementedError

    def to_code(self) -> str:
        raise NotImplementedError

    def to_message(self) -> str:
        raise NotImplementedError


InternalServerErrorCode = "internal_server_error"
InternalServerErrorMessage = (
    "The request could not be completed due to an " "internal server error."
)


class InternalServerAPIError(APIErrorResponse):
    """Raised when an internal server error occurs"""

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_500_INTERNAL_SERVER_ERROR

    def to_code(self) -> str:
        return InternalServerErrorCode

    def to_message(self) -> str:
        return InternalServerErrorMessage


class ValidationAPIError(APIErrorResponse):
    def __init__(self, param: str):
        self.param = param

    def to_status_code(self) -> status:
        return status.HTTP_422_UNPROCESSABLE_ENTITY

    def to_code(self) -> str:
        return "validation_error"

    def to_message(self) -> str:
        return f"Invalid parameter: {self.param}"


class NotFoundAPIError(APIErrorResponse):
    def __init__(self, resource: str):
        self.resource = resource

    def to_status_code(self) -> status:
        return status.HTTP_404_NOT_FOUND

    def to_code(self) -> str:
        return "not_found"

    def to_message(self) -> str:
        return f"{self.resource} not found."


class BadGatewayAPIError(APIErrorResponse):
    """Raised when an error occurs on the upstream server"""

    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_502_BAD_GATEWAY

    def to_code(self) -> str:
        return "bad_gateway"

    def to_message(self) -> str:
        return (
            "The request could not be completed due to the server "
            "receiving an invalid response from an inbound server it "
            "accessed while attempting to fulfill the request. "
            "Check if LLM is running."
        )


class OpenAIError(APIErrorResponse):
    def __init__(self, message):
        self.message = message

    def to_status_code(self) -> status:
        return status.HTTP_502_BAD_GATEWAY

    def to_code(self) -> str:
        return "openai_error"

    def to_message(self) -> str:
        return self.message


class WeaviateConnectionError(APIErrorResponse):
    def __init__(self, message):
        self.message = message

    def to_status_code(self) -> status:
        return status.HTTP_502_BAD_GATEWAY

    def to_code(self) -> str:
        return "weaviate_error"

    def to_message(self) -> str:
        return self.message


class GenericCopilotRuntimeError(APIErrorResponse):
    def __init__(self, message):
        self.message = message

    def to_status_code(self) -> status:
        return status.HTTP_502_BAD_GATEWAY

    def to_code(self) -> str:
        return "copilot_error"

    def to_message(self) -> str:
        return self.message


class AuthorizationMissingAPIError(APIErrorResponse):
    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_401_UNAUTHORIZED

    def to_code(self) -> str:
        return "authorization_missing"

    def to_message(self) -> str:
        return f"Request is missing or has invalid 'Authorization' header."


class InvalidCredentialsAPIError(APIErrorResponse):
    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_401_UNAUTHORIZED

    def to_code(self) -> str:
        return "invalid_credentials"

    def to_message(self) -> str:
        return "Invalid credentials"


class ForbiddenAPIError(APIErrorResponse):
    def __init__(self):
        pass

    def to_status_code(self) -> status:
        return status.HTTP_403_FORBIDDEN

    def to_code(self) -> str:
        return "forbidden"

    def to_message(self) -> str:
        return "User does not have access rights to the content"



from llama_cpp import Llama
from typing import List
import json


class LLamaLLM:
    def __init__(self, model: str, context_size: int) -> None:
        self.context_size = context_size
        self.model = Llama(
            model_path=model, use_mlock=True, n_ctx=context_size, n_gpu_layers=-1
        )

    def generate(
        self,
        prompt: str,
        temperature: float,
        max_tokens: int,
        stop: List[str] = ["###", "User:", "[INST]"],
    ) -> str:
        for token in self.model(
            prompt,
            temperature=temperature,
            max_tokens=max_tokens,
            stream=True,
            stop=stop,
        ):
            output = {"token": {"text": token["choices"][0]["text"]}}
            output = "data:" + json.dumps(output) + "\n"
            output = output.encode("utf-8")
            yield output

    def tokenize(self, text: str) -> List[int]:
        return self.model.tokenize(text.encode("utf-8"))



import os

from opencopilot.domain.errors import APIKeyError
from opencopilot.domain.errors import PromptError
from opencopilot.domain.errors import ModelError
from opencopilot.domain.chat.models.local import LocalLLM


def validate_prompt_and_prompt_file_config(prompt: str, prompt_file: str):
    if prompt and prompt_file:
        raise PromptError(
            "You can only pass either a prompt or a prompt_file argument, not both."
        )
    if not prompt and not prompt_file:
        raise PromptError("You need to pass either a prompt or a prompt_file argument.")

    if prompt_file and not os.path.isfile(prompt_file):
        raise PromptError(
            f"Prompt file '{prompt_file}' does not exist. Please make sure your prompt file path points to a file that exists."
        )


def validate_system_prompt(prompt: str):
    if not "{question}" in prompt:
        raise PromptError(
            f"Template variable '{{question}}' is missing in prompt. Please make sure your prompt file includes all required template variables."
        )
    if not "{history}" in prompt:
        raise PromptError(
            f"Template variable '{{history}}' is missing in prompt. Please make sure your prompt file includes all required template variables."
        )
    if not "{context}" in prompt:
        raise PromptError(
            f"Template variable '{{context}}' is missing in prompt. Please make sure your prompt file includes all required template variables."
        )


def validate_openai_api_key(key: str):
    if not key:
        raise APIKeyError(
            "OpenAI API key is empty or missing. Please add your OpenAI API key either as an environment variable, or an argument to the OpenCopilot() constructor."
        )
    if len(key) != 51 or not key.startswith("sk-"):
        raise APIKeyError(
            "OpenAI API key format is incorrect. Please check that you've entered a correct OpenAI API key."
        )


def validate_local_llm(llm: LocalLLM):
    try:
        llm.get_num_tokens("test")
    except:
        raise ModelError(
            f"Failed to use local LLM.\nMake sure it is running at {llm.llm_url}"
        )


def validate_settings(openai_api_key: str):
    validate_openai_api_key(openai_api_key)



import json
import uuid

import requests

from opencopilot.domain import error_messages
from opencopilot.domain.errors import CopilotIsNotRunningError
from opencopilot.scripts import get_jwt_token

headers = {"accept": "application/json", "Content-Type": "application/json"}
DEFAULT_MESSAGE = "Hi"


def _get_stream(url: str, message: str = DEFAULT_MESSAGE, jwt_token: str = None):
    data = {"message": message}
    if jwt_token:
        headers["Authorization"] = "Bearer " + jwt_token
    s = requests.Session()
    with s.post(url, headers=headers, json=data, stream=True) as resp:
        for line in resp.iter_lines():
            if line:
                yield line


def _process_text(line):
    line = line.decode("utf-8")
    try:
        line = json.loads(line)
        if error := line.get("error"):
            print("ERROR:", error)
            raise Exception("Error in stream")
        return line["text"]
    except:
        return ""


def conversation(
    base_url: str,
    conversation_id: uuid.UUID,
    message: str = DEFAULT_MESSAGE,
):
    url = f"{base_url}/v0/conversations/{conversation_id}"
    try:
        jwt_token = get_jwt_token.execute(base_url)
        if jwt_token:
            headers["Authorization"] = "Bearer " + jwt_token
        data = {"message": message}
        return requests.post(url, json=data, headers=headers)
    except requests.exceptions.ConnectionError:
        raise CopilotIsNotRunningError(
            error_messages.COPILOT_IS_NOT_RUNNING_ERROR.format(copilot_url=url)
        )


def conversation_stream(
    base_url: str,
    conversation_id: uuid.UUID,
    message: str = DEFAULT_MESSAGE,
    stream: bool = False,
):
    url = f"{base_url}/v0/conversations/{conversation_id}/stream"
    try:
        jwt_token = get_jwt_token.execute(base_url)
        output = ""
        for text in _get_stream(url, message=message, jwt_token=jwt_token):
            text = _process_text(text)
            output += text
            if stream:
                print(text, end="", flush=True)
        return output
    except requests.exceptions.ConnectionError:
        print("\n")
        raise CopilotIsNotRunningError(
            error_messages.COPILOT_IS_NOT_RUNNING_ERROR.format(copilot_url=url)
        )


if __name__ == "__main__":
    _result = conversation_stream(
        # TODO: fix base_url
        base_url=f"http://0.0.0.0:3000",
        conversation_id=uuid.uuid4(),
    )
    print("result:", _result)



from typing import Optional

from starlette.requests import Request

from opencopilot.service.middleware.entities import RequestStateKey


def set_state(request: Request, state_key: RequestStateKey, value: any):
    request.state.__setattr__(state_key.value, value)


def get_state(request: Request, state_key: RequestStateKey) -> Optional[any]:
    return getattr(request.state, state_key.value, None)



import logging
from typing import Dict
from typing import Optional
from typing import Union

DEFAULT_LOG_LEVEL = logging.WARNING
LOG_LEVELS: Dict[str, int] = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


def get():
    logging.basicConfig()
    logger = logging.getLogger("OpenCopilot")
    return logger


def set_log_level(log_level: Optional[Union[str, int]] = DEFAULT_LOG_LEVEL):
    logger = logging.getLogger("OpenCopilot")
    logger.setLevel(_convert_log_level(log_level))


def _convert_log_level(log_level: Optional[Union[str, int]]) -> int:
    if not log_level:
        log_level = DEFAULT_LOG_LEVEL
    elif isinstance(log_level, str):
        new_level = LOG_LEVELS.get(log_level)
        if new_level:
            log_level = new_level
        else:
            log_level = DEFAULT_LOG_LEVEL
    return log_level



from typing import List

from filetype import filetype
from langchain.document_loaders import PyPDFLoader
from langchain.schema import Document

from opencopilot.src.utils.loaders.url_loaders.entities import (
    UnsupportedFileTypeException,
)


def execute(file_name: str, url: str) -> List[Document]:
    if not _is_file_type_pdf(file_name):
        raise UnsupportedFileTypeException()
    formatted_docs = []
    loader = PyPDFLoader(file_name)
    docs = loader.load()
    for doc in docs:
        formatted_docs.append(
            Document(
                page_content=doc.page_content,
                metadata={"source": url, "page": doc.metadata.get("page")},
            )
        )
    return formatted_docs


def _is_file_type_pdf(file_name: str) -> bool:
    mime_type: str = ""
    try:
        kind = filetype.guess(file_name)
        mime_type = kind.mime
    except:
        pass
    return mime_type == "application/pdf"



import pickle
from typing import List
from typing import Dict
from pydantic import PrivateAttr

import xxhash
from langchain.embeddings.base import Embeddings
from langchain.embeddings import OpenAIEmbeddings
from openai import OpenAIError

from opencopilot import settings
from opencopilot.domain.errors import OpenAIRuntimeError
from opencopilot.logger import api_logger

logger = api_logger.get()


class CachedEmbeddings:
    embeddings: Embeddings = PrivateAttr()
    _cache: Dict = PrivateAttr()
    _embeddings_cache_filename: str = PrivateAttr()

    def __init__(self, embeddings: Embeddings, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._embeddings = embeddings
        self._cache = {}
        self._embeddings_cache_filename = (
            settings.get().COPILOT_NAME + "_embeddings_cache.pkl"
        )
        self._load_local_cache()

    def embed_documents(self, texts: List[str], **kwargs) -> List[List[float]]:
        return self._embed_documents_cached(texts)

    def embed_query(self, text: str) -> List[float]:
        return self._embeddings.embed_query(text)

    def _embed_documents_cached(self, texts: List[str]) -> List[List[float]]:
        embeddings = []
        try:
            for text in texts:
                text_hash = self._hash(text)
                # pylint: disable-next=no-member
                if embedding := self._cache.get(text_hash):
                    embeddings.append(embedding)
                else:
                    embedding = self._embeddings.embed_documents([text])[0]
                    # pylint: disable-next=no-member
                    self._cache[text_hash] = embedding
                    embeddings.append(embedding)
        except OpenAIError as exc:
            raise OpenAIRuntimeError(exc.user_message)
        return embeddings

    def _hash(self, text) -> str:
        return xxhash.xxh64(text.encode("utf-8")).hexdigest()

    def _load_local_cache(self):
        try:
            # pylint: disable-next=no-member
            with open(self._embeddings_cache_filename, "rb") as f:
                data = pickle.load(f)
                self._cache = data
        except:
            pass

    def save_local_cache(self):
        try:
            with open(self._embeddings_cache_filename, "wb") as f:
                pickle.dump(self._cache, f)
        except Exception as e:
            logger.warning(
                f"Failed to save embeddings cache to {self._embeddings_cache_filename}"
            )


def execute():
    embeddings = settings.get().EMBEDDING_MODEL
    if isinstance(embeddings, str):
        openai_api_base = None
        headers = None
        if settings.get().HELICONE_API_KEY:
            openai_api_base = settings.get().HELICONE_BASE_URL
            headers = {
                "Helicone-Auth": "Bearer " + settings.get().HELICONE_API_KEY,
                "Helicone-Cache-Enabled": "true",
            }
        embeddings = OpenAIEmbeddings(
            disallowed_special=(),
            openai_api_base=openai_api_base,
            headers=headers,
            openai_api_key=settings.get().OPENAI_API_KEY,
        )
    return CachedEmbeddings(embeddings)



import os
from datetime import timedelta
from typing import Callable
from typing import List
from typing import Optional
from typing import Union

import uvicorn
from fastapi import APIRouter
from langchain.chat_models.base import BaseChatModel
from langchain.embeddings.base import Embeddings
from langchain.schema import Document

from opencopilot import exception_utils
from opencopilot import settings
from opencopilot.domain import error_messages
from opencopilot.domain.chat.models.local import LocalLLM
from opencopilot.domain.errors import LogsDirError
from opencopilot.domain.errors import ModelError
from opencopilot.logger import api_logger
from opencopilot.repository.documents import split_documents_use_case
from opencopilot.settings import Settings
from opencopilot.utils.validators import validate_local_llm
from opencopilot.utils.validators import validate_openai_api_key
from opencopilot.utils.validators import validate_prompt_and_prompt_file_config
from opencopilot.utils.validators import validate_system_prompt

ALLOWED_LLM_MODEL_NAMES = ["gpt-3.5-turbo-16k", "gpt-4"]

from opencopilot.analytics import track
from opencopilot.analytics import TrackingEventType

exception_utils.add_copilot_exception_catching()


class OpenCopilot:
    def __init__(
        self,
        prompt: Optional[str] = None,
        prompt_file: Optional[str] = None,
        question_template: Optional[str] = "### Human: {question}",
        response_template: Optional[str] = "### Assistant: {response}",
        openai_api_key: Optional[str] = None,
        copilot_name: str = "default",
        host: str = "127.0.0.1",
        api_port: int = 3000,
        environment: str = "local",
        allowed_origins: str = "*",
        weaviate_url: Optional[str] = None,
        weaviate_read_timeout: int = 120,
        llm: Optional[Union[str, BaseChatModel]] = "gpt-3.5-turbo-16k",
        embedding_model: Optional[Union[str, Embeddings]] = "text-embedding-ada-002",
        max_document_size_mb: int = 50,
        auth_type: Optional[str] = None,
        api_key: Optional[str] = None,
        jwt_client_id: Optional[str] = None,
        jwt_client_secret: Optional[str] = None,
        jwt_token_expiration_seconds: Optional[int] = timedelta(days=1).total_seconds(),
        helicone_api_key: Optional[str] = None,
        helicone_rate_limit_policy: Optional[str] = "3;w=60;s=user",
        logs_dir: Optional[str] = "logs",
        log_level: Optional[Union[str, int]] = None,
    ):
        api_logger.set_log_level(log_level)

        if not openai_api_key:
            openai_api_key = os.getenv("OPENAI_API_KEY")

        tracking_enabled = (
            not os.environ.get("OPENCOPILOT_DO_NOT_TRACK", "").lower() == "true"
        )

        if isinstance(llm, str) or isinstance(embedding_model, str):
            validate_openai_api_key(openai_api_key)
        validate_prompt_and_prompt_file_config(prompt, prompt_file)

        if not prompt:
            with open(prompt_file, "r") as f:
                prompt = f.read()

        validate_system_prompt(prompt)

        if isinstance(llm, str) and llm not in ALLOWED_LLM_MODEL_NAMES:
            raise ModelError(
                error_messages.INVALID_MODEL_ERROR.format(
                    llm_model_name=llm,
                    allowed_model_names=ALLOWED_LLM_MODEL_NAMES,
                )
            )
        if isinstance(llm, LocalLLM):
            validate_local_llm(llm)

        if not logs_dir:
            raise LogsDirError(error_messages.INVALID_LOGS_DIR_ERROR)

        settings.set(
            Settings(
                PROMPT=prompt,
                QUESTION_TEMPLATE=question_template,
                RESPONSE_TEMPLATE=response_template,
                OPENAI_API_KEY=openai_api_key,
                COPILOT_NAME=copilot_name,
                HOST=host,
                API_PORT=api_port,
                ENVIRONMENT=environment,
                ALLOWED_ORIGINS=allowed_origins,
                WEAVIATE_URL=weaviate_url,
                WEAVIATE_READ_TIMEOUT=weaviate_read_timeout,
                LLM=llm,
                EMBEDDING_MODEL=embedding_model,
                MAX_DOCUMENT_SIZE_MB=max_document_size_mb,
                AUTH_TYPE=auth_type,
                API_KEY=api_key,
                JWT_CLIENT_ID=jwt_client_id,
                JWT_CLIENT_SECRET=jwt_client_secret,
                JWT_TOKEN_EXPIRATION_SECONDS=jwt_token_expiration_seconds,
                HELICONE_API_KEY=helicone_api_key,
                HELICONE_RATE_LIMIT_POLICY=helicone_rate_limit_policy,
                TRACKING_ENABLED=tracking_enabled,
                LOGS_DIR=logs_dir,
            )
        )

        self.llm = llm
        self.embedding_model = embedding_model
        self.host = host
        self.api_port = api_port
        self.data_loaders = []
        self.local_files_dirs = []
        self.data_urls = []
        self.local_file_paths = []
        self.documents = []
        self.router = _get_custom_router()

    def __call__(self, *args, **kwargs):
        from .repository.documents import document_loader
        from .repository.documents import document_store
        from opencopilot.repository.documents.document_store import (
            WeaviateDocumentStore,
        )
        from opencopilot.repository.documents.document_store import EmptyDocumentStore
        from .src.utils.loaders import urls_loader

        if (
            self.data_loaders
            or self.local_files_dirs
            or self.local_file_paths
            or self.data_urls
        ):
            self.document_store = WeaviateDocumentStore()
        else:
            self.document_store = EmptyDocumentStore()
        document_store.init_document_store(self.document_store)

        text_splitter = self.document_store.get_text_splitter()
        for data_loader in self.data_loaders:
            documents = data_loader()
            document_chunks = split_documents_use_case.execute(text_splitter, documents)
            self.documents.extend(document_chunks)

        for data_dir in self.local_files_dirs:
            self.documents.extend(
                document_loader.execute(data_dir, False, text_splitter)
            )

        if len(self.data_urls):
            self.documents.extend(
                urls_loader.execute(
                    self.data_urls, text_splitter, settings.get().MAX_DOCUMENT_SIZE_MB
                )
            )

        if self.documents:
            self.document_store.ingest_data(self.documents)

        from .app import app

        app.include_router(self.router)
        track(
            TrackingEventType.COPILOT_START,
            len(self.documents),
            len(self.data_loaders),
            len(self.local_files_dirs),
            len(self.local_file_paths),
            len(self.data_urls),
        )

        uvicorn.run(app, host=self.host, port=self.api_port)

    def data_loader(self, function: Callable[[], Document]):
        self.data_loaders.append(function)

    def add_local_files_dir(self, files_dir: str) -> None:
        self.local_files_dirs.append(files_dir)

    def add_data_urls(self, urls: List[str]) -> None:
        self.data_urls.extend(urls)


def _get_custom_router() -> APIRouter:
    router = APIRouter()
    router.openapi_tags = ["Custom"]
    router.title = "Custom router"
    return router



from datetime import datetime

from opencopilot.domain.chat import is_user_allowed_to_chat_use_case
from opencopilot.domain.chat.entities import MessageModel
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.domain.chat.results import get_gpt_result_use_case
from opencopilot.domain.chat.utils import get_system_message
from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.documents.document_store import DocumentStore
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.error_responses import ForbiddenAPIError

logger = api_logger.get()


async def execute(
    domain_input: UserMessageInput,
    document_store: DocumentStore,
    history_repository: ConversationHistoryRepositoryLocal,
    logs_repository: ConversationLogsRepositoryLocal,
    users_repository: UsersRepositoryLocal,
) -> MessageModel:
    if not is_user_allowed_to_chat_use_case.execute(
        domain_input.conversation_id,
        domain_input.user_id,
        history_repository,
        users_repository,
    ):
        raise ForbiddenAPIError()

    system_message = get_system_message()
    context = []
    if "{context}" in system_message:
        context = document_store.find(domain_input.message)
    message_timestamp = datetime.now().timestamp()
    result = await get_gpt_result_use_case.execute(
        domain_input,
        system_message,
        context,
        logs_repository=logs_repository,
        history_repository=history_repository,
    )

    response_timestamp = datetime.now().timestamp()

    history_repository.save_history(
        domain_input.message,
        result,
        message_timestamp,
        response_timestamp,
        domain_input.conversation_id,
        domain_input.response_message_id,
    )
    users_repository.add_conversation(
        conversation_id=domain_input.conversation_id, user_id=domain_input.user_id
    )
    sources = [document.metadata.get("source") for document in context]

    return MessageModel(
        conversation_id=domain_input.conversation_id, content=result, sources=sources
    )



from opencopilot.domain.chat import chat_delete_use_case
from opencopilot.domain.chat.entities import ChatDeleteInput
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.chat.entities import ChatDeleteRequest
from opencopilot.service.chat.entities import ChatDeleteResponse
from opencopilot.service.error_responses import ForbiddenAPIError
from opencopilot.service.utils import get_uuid


def execute(
    request: ChatDeleteRequest,
    users_repository: UsersRepositoryLocal,
    history_repository: ConversationHistoryRepositoryLocal,
    logs_repository: ConversationLogsRepositoryLocal,
) -> ChatDeleteResponse:
    conversation_id = get_uuid(request.conversation_id, "conversation_id")
    result = chat_delete_use_case.execute(
        data_input=ChatDeleteInput(
            conversation_id=conversation_id, user_id=request.user_id
        ),
        users_repository=users_repository,
        history_repository=history_repository,
        logs_repository=logs_repository,
    )
    if result.response == "NOK":
        raise ForbiddenAPIError()
    return ChatDeleteResponse(response=result.response)



from typing import List
from typing import Tuple

from langchain import PromptTemplate
from langchain.chat_models.base import BaseChatModel
from langchain.schema import Document
from langchain.schema import HumanMessage
from openai import OpenAIError

from opencopilot import settings
from opencopilot.domain.chat import get_token_count_use_case
from opencopilot.domain.chat import utils
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.domain.chat.results import format_context_documents_use_case
from opencopilot.domain.chat.results import get_llm
from opencopilot.domain.errors import OpenAIRuntimeError
from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.utils.callbacks.callback_handler import (
    CustomAsyncIteratorCallbackHandler,
)

logger = api_logger.get()


async def execute(
    domain_input: UserMessageInput,
    system_message: str,
    context: List[Document],
    logs_repository: ConversationLogsRepositoryLocal,
    history_repository: ConversationHistoryRepositoryLocal,
    callback: CustomAsyncIteratorCallbackHandler = None,
) -> str:
    llm = get_llm.execute(domain_input.user_id, streaming=callback is not None)

    history = utils.add_history(
        system_message,
        domain_input.conversation_id,
        history_repository,
    )
    logs_repository.log_history(
        domain_input.conversation_id,
        domain_input.message,
        history.formatted_history,
        domain_input.response_message_id,
        token_count=get_token_count_use_case.execute(history.formatted_history, llm),
    )

    prompt_text = _get_prompt_text(
        domain_input,
        history.template_with_history,
        context,
        llm,
        logs_repository,
    )

    logs_repository.log_prompt_text(
        domain_input.conversation_id,
        domain_input.message,
        prompt_text,
        domain_input.response_message_id,
        token_count=get_token_count_use_case.execute(prompt_text, llm),
    )
    logs_repository.log_prompt_template(
        domain_input.conversation_id,
        domain_input.message,
        system_message,
        domain_input.response_message_id,
        token_count=get_token_count_use_case.execute(
            system_message, llm, is_use_cache=True
        ),
    )

    messages = [HumanMessage(content=prompt_text)]
    try:
        result_message = await llm.agenerate(
            [messages],
            callbacks=[callback] if callback is not None else None,
            stream=callback is not None,
        )
        result = result_message.generations[0][0].text
        return result
    except OpenAIError as exc:
        raise OpenAIRuntimeError(exc.user_message)


def _get_context(
    documents: List[Document],
    llm: BaseChatModel,
) -> Tuple[str, int]:
    while len(documents):
        context = format_context_documents_use_case.execute(documents)
        token_count = get_token_count_use_case.execute(context, llm)
        # Naive solution: leaving 25% for non-context in prompt
        if token_count < (settings.get().get_max_token_count() * 0.75):
            return context, len(documents)
        documents = documents[:-1]
    return "", 0


def _get_prompt_text(
    domain_input: UserMessageInput,
    template_with_history: str,
    context_documents: List[Document],
    llm: BaseChatModel,
    logs_repository: ConversationLogsRepositoryLocal,
) -> str:
    # Almost duplicated with get_local_llm_result_use_case._get_prompt_text
    context, context_documents_count = _get_context(context_documents, llm)
    prompt_text = ""
    if "{context}" in template_with_history:
        prompt = PromptTemplate(
            template=template_with_history, input_variables=["context", "question"]
        )

        prompt_text = prompt.format_prompt(
            **{"context": context, "question": domain_input.message}
        ).to_string()

        if (
            get_token_count_use_case.execute(prompt_text, llm)
            > settings.get().get_max_token_count()
        ):
            prompt_text = ""
        logs_repository.log_context(
            domain_input.conversation_id,
            domain_input.message,
            context_documents[0:context_documents_count],
            domain_input.response_message_id,
            token_count=get_token_count_use_case.execute(context, llm),
        )
    if not prompt_text:
        prompt = PromptTemplate(
            template=template_with_history, input_variables=["context", "question"]
        )
        prompt_text = prompt.format_prompt(
            **{"context": "", "question": domain_input.message}
        ).to_string()
    return prompt_text



from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID


@dataclass(frozen=True)
class ChatConversationsInput:
    user_id: Optional[str] = None


@dataclass(frozen=True)
class ChatConversationsOutput:
    conversations: List[str]


@dataclass(frozen=True)
class MessageModel:
    conversation_id: UUID
    content: str
    sources: List[str]


@dataclass(frozen=True)
class UserMessageInput:
    conversation_id: UUID
    message: str
    response_message_id: str
    user_id: str = None


@dataclass(frozen=True)
class LoadingMessage:
    message: str
    called_copilot: Optional[str]

    def to_dict(self) -> Dict:
        result = {"message": self.message}
        if self.called_copilot:
            result["called_copilot"] = self.called_copilot
        return result

    @staticmethod
    def from_dict(loading_message: Dict):
        return LoadingMessage(
            message=loading_message.get("message") or "",
            called_copilot=loading_message.get("called_copilot") or None,
        )


@dataclass(frozen=True)
class StreamingChunk:
    conversation_id: UUID
    text: str
    sources: List[str]
    error: Optional[str] = None
    loading_message: Optional[LoadingMessage] = None

    def to_dict(self) -> Dict:
        result = {"text": self.text}
        if self.error:
            result["error"] = self.error
        if self.loading_message:
            result["loading_message"] = self.loading_message.to_dict()
        return result


@dataclass(frozen=True)
class ChatDeleteInput:
    conversation_id: UUID
    user_id: Optional[str] = None


@dataclass(frozen=True)
class ChatDeleteOutput:
    response: str



from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

from opencopilot.domain.debug.entities import MessageDebugResult
from opencopilot.domain.debug.entities import TextWithTokens
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.error_responses import ForbiddenAPIError


def execute(
    conversation_id: UUID,
    message_id: str,
    history_repository: ConversationHistoryRepositoryLocal,
    logs_repository: ConversationLogsRepositoryLocal,
    users_repository: UsersRepositoryLocal,
    user_id: Optional[str] = None,
) -> MessageDebugResult:
    conversations = users_repository.get_conversations(user_id)
    if str(conversation_id) not in conversations:
        raise ForbiddenAPIError()

    history = history_repository.get_history(conversation_id)
    logs_history = logs_repository.get_logs_by_message(conversation_id, message_id)
    return MessageDebugResult(
        prompt_template=_get_logs_history_value(logs_history, "prompt_template"),
        data_sources=None,
        user_question=_get_history_value(history, message_id, "prompt"),
        context=_get_logs_history_value(logs_history, "context"),
        chat_history=_get_logs_history_value(logs_history, "history"),
        full_prompt=_get_logs_history_value(logs_history, "prompt_text"),
        llm_response=_get_history_value(history, message_id, "response"),
    )


def _get_history_value(
    history: List[Dict],
    message_id: str,
    value_key: str,
) -> Optional[TextWithTokens]:
    for h in history:
        if h.get("response_message_id") == message_id:
            return TextWithTokens(
                text=h.get(value_key),
                token_count=None,
            )

    return None


def _get_logs_history_value(
    history: List[Dict],
    value_key: str,
) -> Optional[TextWithTokens]:
    for h in history:
        if value := h.get(value_key):
            return TextWithTokens(
                text=value,
                token_count=h.get("token_count"),
            )
    return None



from dataclasses import dataclass
from uuid import UUID

from opencopilot import settings
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)


@dataclass(frozen=True)
class History:
    template_with_history: str
    formatted_history: str


def add_history(
    template: str,
    conversation_id: UUID,
    history_repository: ConversationHistoryRepositoryLocal,
) -> History:
    history = history_repository.get_prompt_history(
        conversation_id, settings.get().PROMPT_HISTORY_INCLUDED_COUNT
    )
    history = history.replace("{", "{{").replace("}", "}}")
    return History(
        template_with_history=template.replace("{history}", history, 1),
        formatted_history=history,
    )


def get_system_message() -> str:
    return settings.get().PROMPT



import time
import uuid

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from opencopilot.logger import api_logger
from opencopilot.service.middleware import util
from opencopilot.service.middleware.entities import RequestStateKey

logger = api_logger.get()


class RequestEnrichmentMiddleware(BaseHTTPMiddleware):
    def __init__(
        self,
        app: ASGIApp,
    ) -> None:
        super().__init__(app)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        start = time.time()
        request_id = uuid.uuid4()

        request_source = request.headers.get("source")
        request_origin = request.headers.get("origin")
        request_user_agent = request.headers.get("user-agent")
        ip_address = request.headers.get("x-forwarded-for", "").split(",")[0]

        util.set_state(request, RequestStateKey.REQUEST_ID, request_id)
        util.set_state(request, RequestStateKey.REQUEST_SOURCE, request_source)
        util.set_state(request, RequestStateKey.REQUEST_ORIGIN, request_origin)
        util.set_state(request, RequestStateKey.REQUEST_USER_AGENT, request_user_agent)
        util.set_state(request, RequestStateKey.IP_ADDRESS, ip_address)

        time_elapsed = time.time() - start
        if time_elapsed > 1:
            logger.debug(f"Middleware Enrichment before request time: {time_elapsed}")

        return await call_next(request)



import json
from typing import List

from langchain.document_loaders import TextLoader
from langchain.schema import Document

from opencopilot.src.utils.loaders.url_loaders.entities import (
    UnsupportedFileTypeException,
)


def execute(file_name: str, url: str) -> List[Document]:
    if not _is_file_type_json(file_name):
        raise UnsupportedFileTypeException()
    loader = TextLoader(file_name)
    documents = loader.load()
    formatted_documents = []
    for document in documents:
        metadata = document.metadata or {}
        metadata["source"] = url
        formatted_documents.append(
            Document(page_content=document.page_content, metadata=metadata)
        )
    return formatted_documents


def _is_file_type_json(file_name: str) -> bool:
    try:
        with open(file_name, "r") as f:
            content = f.read()
        json.loads(content)
    except:
        return False
    return True



import argparse
import json
from dataclasses import dataclass
from typing import List
import warnings

from tqdm import tqdm
from langchain import PromptTemplate
from langchain.chat_models import ChatOpenAI
from langchain.schema import SystemMessage
import settings
from src.repository.documents import document_store

from src.eval.entities import RetrievalExample, RetrievalDataset, DocumentID

PROMPT_TEMPLATE = """"
You are a professional game developer with 20 years of experience.
Your goal is to help other developers integrate Ready Player Me Avatar SDK.

Please generate 5 questions from given url and context which developers might ask. Please think this through step by step so we come to factual and helpful questions.
Avoid generating bad questions.

GOOD QUESTIONS:
What do I need to do to use Ready Player Me avatars for commercial use?
Can I control the logging feature of the avatar loading process?
What should I do to avoid errors when uploading to Mixamo?
How can I add animations to Ready Player Me avatars?
How can I optimize the avatars for my game?

BAD QUESTIONS:
What does the WebViewActivity.kt file do?
What is the 'GetAvatarCacheData' function used for?
What is the purpose of the AvatarRenderLoader class?
What is the ReadyPlayerMeComponent?
What is the purpose of the 'FDownloadImageCompleted& OnCompleted' parameter in the 'Load' function?

Output questions in JSON list format like [{{"query": "First question?", "answer": "First answer"}},...].

URL: {url}
CONTEXT: {context}
"""

PROMPT = PromptTemplate(input_variables=["url", "context"], template=PROMPT_TEMPLATE)


@dataclass(frozen=True)
class DocumentQueries:
    source_document: DocumentID
    queries: List[str]


def main(dataset_path: str) -> None:
    print("question generation starting!")
    llm = ChatOpenAI(temperature=0.3, model_name="gpt-4")
    # TODO: document_store.init_document_store()
    store = document_store.get_document_store()
    documents = store.load_documents()
    generated_questions = []
    for document in tqdm(documents):
        response = llm(
            [
                SystemMessage(
                    content=PROMPT_TEMPLATE.format(
                        url=document.metadata["source"], context=document.page_content
                    )
                )
            ]
        )
        if questions := _parse_questions(response.content, document.metadata["source"]):
            generated_questions.append(questions)
    _save_dataset(generated_questions, dataset_path)


def _save_dataset(doc_queries_list: List[DocumentQueries], dataset_path: str):
    examples = []
    for doc_queries in doc_queries_list:
        for query in doc_queries.queries:
            retrieval_example = RetrievalExample(
                query=query, documents=[doc_queries.source_document]
            )

            examples.append(retrieval_example)

    dataset = RetrievalDataset(examples=examples)

    # Writing to JSON
    with open(dataset_path, "w") as file:
        json.dump(dataset.to_dict(), file, indent=4)


def _parse_questions(
    llm_response: str, source_document: DocumentID
) -> List[DocumentQueries]:
    queries = []
    try:
        json_llm_response = json.loads(llm_response)
        for response in json_llm_response:
            queries.append(response["query"])
    except:
        warnings.warn(
            f"Could not parse LLM response for document {source_document}: {llm_response}"
        )
    return DocumentQueries(queries=queries, source_document=source_document)


if __name__ == "__main__":
    assert settings.get().OPENAI_API_KEY, "OpenAI API key needs to be present"
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dataset_path",
        type=str,
        required=False,
        default="../copilots/rpm/eval_data/retrieval_auto.json",
    )
    args = parser.parse_args()
    main(args.dataset_path)



from opencopilot.authorization import create_access_token
from opencopilot.service.authorization.entities import TokenRequest, TokenResponse


def execute(request: TokenRequest) -> TokenResponse:
    token = create_access_token.execute(
        request.client_id, request.client_secret, request.user_id
    )
    return TokenResponse(response="OK", token=token)



import os
from opencopilot import settings
from opencopilot.settings import Settings


def set_default_settings(name: str = "script"):
    settings.set(
        Settings(
            COPILOT_NAME=name,
            HOST="127.0.0.1",
            API_PORT=3000,
            ENVIRONMENT=name,
            ALLOWED_ORIGINS="*",
            WEAVIATE_URL="http://localhost:8080/",
            WEAVIATE_READ_TIMEOUT=120,
            LLM="gpt-4",
            EMBEDDING_MODEL="text-embedding-ada-002",
            OPENAI_API_KEY=os.getenv("OPENAI_API_KEY"),
            MAX_DOCUMENT_SIZE_MB=1,
            AUTH_TYPE=None,
            API_KEY="",
            JWT_CLIENT_ID="",
            JWT_CLIENT_SECRET="",
            JWT_TOKEN_EXPIRATION_SECONDS=1,
            HELICONE_API_KEY="",
            HELICONE_RATE_LIMIT_POLICY="",
            TRACKING_ENABLED=os.environ.get("OPENCOPILOT_DO_NOT_TRACK", "").lower()
            != "True",
        )
    )



from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from dataclasses import dataclass


@dataclass
class PydanticErrorResponseModel:
    code: str
    message: Union[str, Dict]


def _value_error_number_not_le_ge(error: Dict) -> Optional[PydanticErrorResponseModel]:
    loc = _get_loc(error)
    if not loc:
        return
    message: str = error.get("msg")
    message = message.replace("ensure this value is", "must be")
    return PydanticErrorResponseModel(code="value_error", message=f"{loc} {message}.")


def _value_error_const(error: Dict) -> Optional[PydanticErrorResponseModel]:
    # Same as _type_error_enum but for POST requests
    loc = _get_loc(error)
    if not loc:
        return
    given_value = error.get("ctx", {}).get("given")
    allowed_values = error.get("ctx", {}).get("permitted")
    allowed_values = [allowed_value.value for allowed_value in allowed_values]
    return PydanticErrorResponseModel(
        code="invalid_enumeration",
        message=f"{loc} value {given_value} is not a valid enumeration "
        f"member. Allowed values: {allowed_values}.",
    )


def _value_error_missing(error: Dict) -> Optional[PydanticErrorResponseModel]:
    loc = _get_loc(error)
    if not loc:
        return
    return PydanticErrorResponseModel(
        code="parameter_missing", message=f"Missing required parameter: {loc}."
    )


def _type_error_enum(error: Dict) -> Optional[PydanticErrorResponseModel]:
    # Same as _value_error_const but for Query params
    loc = _get_loc(error)
    if not loc:
        return
    message: str = error.get("msg")
    allowed_values = error.get("ctx", {}).get("enum_values")
    allowed_values = [allowed_value.value for allowed_value in allowed_values]
    message = message.replace("value", loc)
    message = message[: message.find(";")]
    return PydanticErrorResponseModel(
        code="invalid_enumeration",
        message=f"{message}. Allowed values: {allowed_values}.",
    )


def _type_error_generic(error: Dict) -> Optional[PydanticErrorResponseModel]:
    loc = _get_loc(error)
    if not loc:
        return
    message: str = error.get("msg")
    message = message.replace("value", loc)
    return PydanticErrorResponseModel(code="invalid_type", message=f"{message}.")


def _type_error_str(error: Dict) -> Optional[PydanticErrorResponseModel]:
    loc = _get_loc(error)
    if not loc:
        return
    message: str = error.get("msg")
    return PydanticErrorResponseModel(
        code="invalid_type", message=f"{message} for {loc}."
    )


def _get_loc(error) -> Optional[str]:
    # Returns invalid parameter name
    for loc in reversed(error.get("loc")):
        if isinstance(loc, str):
            return loc


expected_errors = {
    "value_error.number.not_le": _value_error_number_not_le_ge,
    "value_error.number.not_ge": _value_error_number_not_le_ge,
    "value_error.const": _value_error_const,
    "value_error.missing": _value_error_missing,
    "type_error.enum": _type_error_enum,
    "type_error.integer": _type_error_generic,
    "type_error.float": _type_error_generic,
    "type_error.list": _type_error_generic,
    "type_error.bool": _type_error_generic,
    "type_error.str": _type_error_str,
}


def execute(error: List[Dict]) -> Optional[PydanticErrorResponseModel]:
    error = error[0]
    try:
        code = error.get("type", "")
        if code in expected_errors:
            if response := expected_errors.get(code)(error):
                return response
    except:
        pass
    try:
        code = error.get("type", "unprocessable_entity").lower().replace(".", "_")
        return PydanticErrorResponseModel(code=code, message=error)
    except:
        pass
    return



import logging
import sys

from colorama import Fore
from colorama import Style

from opencopilot.domain.errors import CopilotConfigurationError
from opencopilot.domain.errors import CopilotRuntimeError
from opencopilot.logger import api_logger


def add_copilot_exception_catching():
    def on_crash(exctype, value, traceback):
        # "exctype" is the class of the exception raised
        # "value" is the instance
        # "traceback" is the object containing what python needs to print
        # logger = logging.getLogger("OpenCopilot")
        logger = api_logger.get()
        if logger.level >= logging.WARNING:
            if issubclass(exctype, CopilotConfigurationError) or issubclass(
                exctype, CopilotRuntimeError
            ):
                # Instead of the stack trace, we print an error message to stderr
                logger.error(f"{Fore.RED}{exctype.__name__}{Style.RESET_ALL}: {value}")
            else:
                # sys.__excepthook__ is the default excepthook that prints the stack trace
                # so we use it directly if we want to see it
                sys.__excepthook__(exctype, value, traceback)
        else:
            sys.__excepthook__(exctype, value, traceback)

    # Now we replace the default excepthook by our own
    sys.excepthook = on_crash



from typing import List
from opencopilot import settings

from langchain import PromptTemplate
from langchain.chat_models import ChatOpenAI
from langchain.evaluation.qa import QAEvalChain


from opencopilot.eval.entities import (
    EndToEndExample,
    EndToEndResult,
    EndToEndDataset,
    EndToEndSingleEvaluation,
    EndToEndSummaryEvaluation,
)

PROMPT_TEMPLATE = """You are a teacher grading a quiz.
You are given a question, the student's answer, and the true answer, and are asked to grade the student answer as either A, B, C, D, or F.

Example Format:
QUESTION: question here
STUDENT ANSWER: student's answer here
TRUE ANSWER: true answer here
GRADE: A, B, C, D or F here

Grade the student answers based ONLY on their factual accuracy, helpfulness and completeness. Ignore differences in punctuation and phrasing between the student answer and true answer. It is OK if the student answer contains more information than the true answer, as long as it does not contain any conflicting statements. Begin! 

QUESTION: {query}
STUDENT ANSWER: {result}
TRUE ANSWER: {answer}
GRADE:"""
PROMPT = PromptTemplate(
    input_variables=["query", "result", "answer"], template=PROMPT_TEMPLATE
)


def get_summary_evaluation(
    evaluations: List[EndToEndSingleEvaluation],
) -> EndToEndSummaryEvaluation:
    """Roll up a list of single evaluations into a summary evaluation."""
    evaluations_count = len(evaluations)
    total_score = 0
    rubric = {"A": 1.0, "B": 0.75, "C": 0.5, "D": 0.25, "F": 0.0}
    for evaluation in evaluations:
        grade = rubric.get(evaluation.evaluation)
        if grade:  # TODO - failures of grading are implicitly scored as zero
            total_score += grade
    return EndToEndSummaryEvaluation(
        evaluations_count=evaluations_count,
        evaluations_score=round((total_score / evaluations_count) * 100, 4),
        single_evaluations=evaluations,
    )


def evaluate_endtoend_single(
    example: EndToEndExample, prediction: EndToEndResult
) -> EndToEndSingleEvaluation:
    llm = ChatOpenAI(temperature=0, model_name="gpt-4")

    # patch it for context token limit - TODO is this still needed?
    settings.get().MODEL = "gpt-4"

    eval_chain = QAEvalChain.from_llm(llm=llm, prompt=PROMPT)
    eval_example = {"query": example.query, "answer": example.answer}
    eval_prediction = {"result": prediction.answer}

    graded_output = eval_chain.evaluate([eval_example], [eval_prediction])[0]
    print("graded_output:", graded_output)

    return EndToEndSingleEvaluation(evaluation=graded_output["results"])


def evaluate_endtoend_dataset(
    dataset: EndToEndDataset, predictions: List[EndToEndResult]
) -> EndToEndSummaryEvaluation:
    assert len(dataset.examples) == len(
        predictions
    ), "Dataset and predictions must be the same length"

    # Convert the results into our own format
    evaluations = [
        evaluate_endtoend_single(example, prediction)
        for example, prediction in zip(dataset.examples, predictions)
    ]

    summary_evaluation = get_summary_evaluation(evaluations)
    return summary_evaluation



import os
import urllib.request
from typing import Callable
from typing import List

from langchain.schema import Document
from langchain.text_splitter import TextSplitter

from opencopilot import settings
from opencopilot.logger import api_logger
from opencopilot.repository.documents import split_documents_use_case
from opencopilot.src.utils.loaders.url_loaders import csv_loader_use_case
from opencopilot.src.utils.loaders.url_loaders import html_loader_use_case
from opencopilot.src.utils.loaders.url_loaders import json_loader_use_case
from opencopilot.src.utils.loaders.url_loaders import pdf_loader_use_case
from opencopilot.src.utils.loaders.url_loaders import xls_loader_use_case

logger = api_logger.get()

loaders: List[Callable[[str, str], List[Document]]] = [
    pdf_loader_use_case.execute,
    csv_loader_use_case.execute,
    xls_loader_use_case.execute,
    json_loader_use_case.execute,
    html_loader_use_case.execute,
]


def execute(
    urls: List[str], text_splitter: TextSplitter, max_document_size_mb: int
) -> List[Document]:
    documents: List[Document] = []
    for url in urls:
        documents.extend(_load_url(url, max_document_size_mb))
    return split_documents_use_case.execute(text_splitter, documents)


def _load_url(url: str, max_document_size_mb: int) -> List[Document]:
    docs: List[Document] = []
    try:
        file_name, headers = urllib.request.urlretrieve(url)
        if (file_size := _get_file_size(file_name)) > max_document_size_mb:
            logger.warning(
                f"Document {url} too big ({file_size} > {settings.get().MAX_DOCUMENT_SIZE_MB}), skipping."
            )
            return []
        for loader in loaders:
            try:
                new_docs = loader(file_name, url)
                if new_docs:
                    docs.extend(new_docs)
                    break
            except Exception as e:
                pass
    except Exception as e:
        pass
    if not docs:
        logger.warning(f"Failed to scrape the contents from {url}")
    return docs


def _get_file_size(file_path):
    size_in_bytes = os.path.getsize(file_path)
    size_in_mb = size_in_bytes / 1024 / 1024
    return size_in_mb



import json
from http.client import responses
from typing import Optional

from fastapi import Request
from fastapi import status
from fastapi.encoders import jsonable_encoder
from fastapi.exception_handlers import http_exception_handler
from fastapi.exception_handlers import request_validation_exception_handler
from fastapi.exceptions import RequestValidationError
from starlette.responses import JSONResponse

from opencopilot.service.error_responses import APIErrorResponse
from opencopilot.service.error_responses import InternalServerAPIError
from opencopilot.utils import http_headers, format_pydantic_validation_error
from opencopilot.utils.format_pydantic_validation_error import (
    PydanticErrorResponseModel,
)


async def custom_exception_handler(request: Request, error: Exception):
    if not isinstance(error, APIErrorResponse):
        error = InternalServerAPIError()
    return await http_headers.add_response_headers(
        JSONResponse(
            status_code=error.to_status_code(),
            content=jsonable_encoder(
                {
                    "response": "NOK",
                    "error": {
                        "status_code": error.to_status_code(),
                        "code": error.to_code(),
                        "message": error.to_message(),
                    },
                }
            ),
        ),
        None,
    )


def process_for_multipart_form_error(response_json: dict):
    if (
        response_json.get("error", {})
        .get("message", "")
        .find("Missing boundary in multipart..")
        > -1
    ):
        response_json["error"][
            "message"
        ] = "The request is missing a file to be uploaded. Please attach a file in the body of the request"
    return response_json


async def custom_http_exception_handler(request, exc):
    response = await http_exception_handler(request, exc)
    try:
        _js = json.loads(response.body)
        if "response" not in _js and "code" not in _js and "status_code" not in _js:
            _js["response"] = "NOK"
            message = _js.pop("detail") + "."
            _js["error"] = {
                "status_code": response.status_code,
                "code": responses[response.status_code].lower().replace(" ", "_"),
                "message": message,
            }
            _js = process_for_multipart_form_error(_js)
            return JSONResponse(
                status_code=response.status_code, content=jsonable_encoder(_js)
            )
    except:
        pass
    return response


async def validation_exception_handler(request: Request, exc: RequestValidationError):
    response = await request_validation_exception_handler(request, exc)
    error: Optional[PydanticErrorResponseModel]
    try:
        if error := format_pydantic_validation_error.execute(exc.errors()):
            return JSONResponse(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                content=jsonable_encoder(
                    {
                        "response": "NOK",
                        "error": {
                            "status_code": status.HTTP_422_UNPROCESSABLE_ENTITY,
                            "code": error.code,
                            "message": error.message,
                        },
                    }
                ),
            )
    except:
        pass
    return response



import json
import os
import csv
from dataclasses import dataclass
from typing import List
import re

from langchain.document_loaders import CSVLoader
from langchain.document_loaders import PyPDFLoader
from langchain.document_loaders import TextLoader
from langchain.document_loaders import UnstructuredExcelLoader


@dataclass(frozen=True)
class ValidateFileResult:
    is_valid: bool
    error_message: str = None


@dataclass(frozen=True)
class ValidateCopilotResult:
    is_valid: bool
    error_results: List[ValidateFileResult]


def execute(relative_directory_path: str) -> ValidateCopilotResult:
    file_paths = [
        os.path.join(dp, f)
        for dp, dn, filenames in os.walk(relative_directory_path)
        for f in filenames
    ]
    error_results = _validate_required_files(relative_directory_path)
    error_results += _validate_prompt_template(relative_directory_path)
    for file_path in file_paths:
        if "/.git/" not in file_path:
            result = _validate_file(file_path)
            if not result.is_valid:
                error_results.append(result)

    if error_results:
        return ValidateCopilotResult(is_valid=False, error_results=error_results)
    return ValidateCopilotResult(is_valid=True, error_results=[])


def _validate_required_files(relative_directory_path: str) -> List[ValidateFileResult]:
    result = []
    files_to_check = [
        "prompts/prompt_template.txt",
        "prompts/prompt_configuration.json",
    ]
    for file in files_to_check:
        file_path = os.path.join(relative_directory_path, file)
        if not os.path.exists(file_path):
            result.append(
                ValidateFileResult(is_valid=False, error_message=f"Missing {file} file")
            )
    return result


def _validate_prompt_template(relative_directory_path: str) -> List[ValidateFileResult]:
    required_params = ["{context}", "{history}", "{question}"]
    file_path = os.path.join(relative_directory_path, "prompts/prompt_template.txt")
    if os.path.exists(file_path):
        content = _read_file(file_path)
        for param in required_params:
            count = _find_word_count_in_string(param, content)
            if count == 0:
                return [
                    ValidateFileResult(
                        is_valid=False,
                        error_message="Prompt template (prompts/prompt_template.txt) is missing {context}"
                        ", {history} or {question}",
                    )
                ]
            elif count > 1:
                return [
                    ValidateFileResult(
                        is_valid=False,
                        error_message="Prompt template (prompts/prompt_template.txt) is having duplicate of {context}, "
                        "{history} or {question}",
                    )
                ]
    return []


def _find_word_count_in_string(word: str, content: str) -> int:
    count = 0
    for _ in re.finditer(word, content):
        count += 1
    return count


def _validate_file(file_path: str) -> ValidateFileResult:
    ext = os.path.splitext(file_path)[1]
    if ext == ".json":
        try:
            loader = TextLoader(file_path)
            loader.load()
            content = _read_file(file_path)
            json.loads(content)
        except:
            return ValidateFileResult(
                is_valid=False, error_message=f"Cannot load json file: {file_path}"
            )
    elif ext == ".csv":
        try:
            with open(file_path, newline="") as csvfile:
                dialect = csv.Sniffer().sniff(csvfile.read(1024))

            loader = CSVLoader(
                file_path=file_path,
                csv_args={
                    "delimiter": dialect.delimiter,
                },
            )
            loader.load()
        except:
            return ValidateFileResult(
                is_valid=False, error_message=f"Cannot csv json file: {file_path}"
            )
    elif ext == ".pdf":
        try:
            loader = PyPDFLoader(file_path)
            loader.load()
        except:
            return ValidateFileResult(
                is_valid=False, error_message=f"Cannot load pdf file: {file_path}"
            )
    elif file_path.endswith(".xls") or file_path.endswith(".xlsx"):
        try:
            loader = UnstructuredExcelLoader(file_path)
            loader.load()
        except:
            return ValidateFileResult(
                is_valid=False, error_message=f"Cannot load excel file: {file_path}"
            )
    return ValidateFileResult(is_valid=True)


def _read_file(file_path: str) -> str:
    with open(file_path, "r", encoding="utf-8") as file:
        return file.read()



from opencopilot.domain.chat import get_chat_history_use_case
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.chat.entities import ChatHistoryRequest
from opencopilot.service.chat.entities import ChatHistoryResponse
from opencopilot.service.error_responses import ForbiddenAPIError
from opencopilot.service.utils import get_uuid


async def execute(
    request: ChatHistoryRequest,
    history_repository: ConversationHistoryRepositoryLocal,
    users_repository: UsersRepositoryLocal,
) -> ChatHistoryResponse:
    conversation_id = get_uuid(request.conversation_id, "conversation_id")
    messages = await get_chat_history_use_case.execute(
        conversation_id=conversation_id,
        user_id=request.user_id,
        history_repository=history_repository,
        users_repository=users_repository,
    )
    if not messages:
        raise ForbiddenAPIError()
    return ChatHistoryResponse(
        response="OK",
        conversation_id=str(conversation_id),
        messages=messages,
    )



from typing import List
from typing import Optional

import tqdm
import weaviate
from langchain.schema import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.text_splitter import TextSplitter
from langchain.vectorstores import Weaviate
from requests.exceptions import InvalidSchema
from requests.exceptions import MissingSchema
from weaviate.exceptions import WeaviateBaseError

from opencopilot import settings
from opencopilot.domain import error_messages
from opencopilot.domain.errors import WeaviateRuntimeError
from opencopilot.utils import get_embedding_model_use_case
from opencopilot.utils.get_embedding_model_use_case import CachedEmbeddings


class DocumentStore:
    document_embed_model = "text-embedding-ada-002"
    document_chunk_size = 2000

    def __init__(self):
        embeddings = settings.get().EMBEDDING_MODEL
        if not isinstance(embeddings, str):
            # smaller chunks if not using OpenAI
            self.document_chunk_size = 500

    def get_embeddings_model(self) -> CachedEmbeddings:
        return get_embedding_model_use_case.execute()

    def get_text_splitter(self) -> TextSplitter:
        return RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            chunk_size=self.document_chunk_size,
            model_name=self.document_embed_model,
            disallowed_special=(),
        )

    def ingest_data(self, documents: List[Document]):
        pass

    def find(self, query: str, **kwargs) -> List[Document]:
        return []


class WeaviateDocumentStore(DocumentStore):
    ingest_batch_size = 100

    weaviate_index_name = "LangChain"  # TODO: Weaviate specific?

    def __init__(self):
        super().__init__()
        self.documents = []
        self.embeddings = self.get_embeddings_model()
        self.weaviate_client = self._get_weaviate_client()
        self.vector_store = self._get_vector_store()

    def _get_weaviate_client(self):
        try:
            if url := settings.get().WEAVIATE_URL:
                return weaviate.Client(
                    url=url,
                    timeout_config=(10, settings.get().WEAVIATE_READ_TIMEOUT),
                )
            else:
                return weaviate.Client(
                    timeout_config=(10, settings.get().WEAVIATE_READ_TIMEOUT),
                    embedded_options=weaviate.embedded.EmbeddedOptions(
                        port=8080,
                        hostname="localhost",
                    ),
                )
        except WeaviateBaseError as exc:
            raise WeaviateRuntimeError(
                exc.message + error_messages.WEAVIATE_ERROR_EXTRA
            )
        except MissingSchema:
            raise WeaviateRuntimeError(
                error_messages.WEAVIATE_INVALID_URL.format(
                    weaviate_url=settings.get().WEAVIATE_URL or "http://localhost:8080"
                )
            )
        except InvalidSchema:
            raise WeaviateRuntimeError(
                error_messages.WEAVIATE_INVALID_URL.format(
                    weaviate_url=settings.get().WEAVIATE_URL or "http://localhost:8080"
                )
            )

    def _get_vector_store(self):
        metadatas = [d.metadata for d in self.documents]
        attributes = list(metadatas[0].keys()) if metadatas else ["source"]
        return Weaviate(
            self.weaviate_client,
            index_name=self.weaviate_index_name,
            text_key="text",
            embedding=self.embeddings,
            attributes=attributes,
            by_text=False,
        )

    def ingest_data(self, documents: List[Document]):
        try:
            self.documents = documents
            batch_size = self.ingest_batch_size
            print(
                f"Got {len(documents)} documents, embedding with batch "
                f"size: {batch_size}"
            )
            self.weaviate_client.schema.delete_all()

            for i in tqdm.tqdm(
                range(0, int(len(documents) / batch_size) + 1), desc="Embedding.."
            ):
                batch = documents[i * batch_size : (i + 1) * batch_size]
                self.vector_store.add_documents(batch)

            self.embeddings.save_local_cache()
            self.vector_store = self._get_vector_store()
        except WeaviateBaseError as exc:
            raise WeaviateRuntimeError(
                exc.message + error_messages.WEAVIATE_ERROR_EXTRA
            )

    def find(self, query: str, **kwargs) -> List[Document]:
        try:
            kwargs["k"] = kwargs.get("k", settings.get().MAX_CONTEXT_DOCUMENTS_COUNT)
            documents = self.vector_store.similarity_search(query, **kwargs)
            return documents
        except WeaviateBaseError as exc:
            raise WeaviateRuntimeError(
                exc.message + error_messages.WEAVIATE_ERROR_EXTRA
            )

    def find_by_source(self, source: str, **kwargs) -> List[Document]:
        try:
            query = (
                self._get_weaviate_client()
                .query.get(self.weaviate_index_name, ["text", "source"])
                .with_additional(["id"])
                .with_where(
                    {"path": ["source"], "operator": "Like", "valueString": source}
                )
            )
            result = (
                query.do()
                .get("data", {})
                .get("Get", {})
                .get(self.weaviate_index_name, [])
            )
            docs = []
            for res in result:
                text = res.pop("text")
                docs.append(Document(page_content=text, metadata=res))
            return docs
        except WeaviateBaseError as exc:
            raise WeaviateRuntimeError(
                exc.message + error_messages.WEAVIATE_ERROR_EXTRA
            )

    def get_all(self) -> List[Document]:
        try:
            client = self._get_weaviate_client()
            batch_size = 200
            cursor = None
            all_results = []

            query = (
                client.query.get(self.weaviate_index_name, ["text", "source"])
                .with_additional(["id"])
                .with_limit(batch_size)
            )

            while True:
                results = query.with_after(cursor).do() if cursor else query.do()
                current_results = results["data"]["Get"].get(
                    self.weaviate_index_name, []
                )
                if not current_results:
                    break
                all_results.extend(current_results)
                cursor = current_results[-1]["_additional"]["id"]

            docs = [
                Document(page_content=res.pop("text"), metadata=res)
                for res in all_results
            ]
            return docs
        except WeaviateBaseError as exc:
            raise WeaviateRuntimeError(
                exc.message + error_messages.WEAVIATE_ERROR_EXTRA
            )


class EmptyDocumentStore(DocumentStore):
    pass


DOCUMENT_STORE = Optional[DocumentStore]


def init_document_store(document_store: DocumentStore):
    global DOCUMENT_STORE
    DOCUMENT_STORE = document_store


def get_document_store() -> DocumentStore:
    return DOCUMENT_STORE



from typing import List

from langchain.schema import Document
from langchain.text_splitter import TextSplitter


def execute(text_splitter: TextSplitter, documents: List[Document]) -> List[Document]:
    document_chunks = []
    for document in documents:
        for chunk in text_splitter.split_text(document.page_content):
            document_chunks.append(
                Document(page_content=chunk, metadata=document.metadata)
            )
    return document_chunks



import uuid
import enum
import xxhash
import platform
import subprocess
from subprocess import CalledProcessError
import importlib.metadata as importlib_metadata

import segment.analytics as segment_analytics


from . import settings
from .settings import Settings


SEGMENT_WRITE_KEY = "iEkciV19ocu9abTCzjbvjPsCKzOJF30S"
segment_analytics.write_key = SEGMENT_WRITE_KEY


class TrackingEventType(enum.Enum):
    COPILOT_START = 1
    CHAT_MESSAGE = 2


def get_opencopilot_version():
    package_name = "opencopilot-ai"
    declared_version = None

    try:
        declared_version = importlib_metadata.version(package_name)
    except importlib_metadata.PackageNotFoundError:
        pass  # Not installed

    # Run `pip freeze` to detect local install and get commit hash
    try:
        pip_freeze_output = subprocess.check_output(["pip", "freeze"]).decode("utf-8")
        matching_lines = [
            l for l in pip_freeze_output.split("\n") if "opencopilot" in l
        ]

        for line in matching_lines:
            if line.startswith(package_name):
                # Normal PyPI install
                declared_version = line.split("==")[1]
                break
            elif line.startswith(f"-e git+") and "egg=opencopilot_ai" in line:
                # Local install from version-controlled repo
                declared_version = line.split("-e ")[1]
                break
            elif "opencopilot-ai" in line or "opencopilot_ai" in line:
                # Local install from non-version-controlled directory
                declared_version = line
    except CalledProcessError:
        pass

    return declared_version


def hashed(s: str):
    return xxhash.xxh64(s.encode("utf-8")).hexdigest()


def get_hashed_user_id():
    """Returns a hashed unique identifier for the user."""
    mac_address: int = uuid.getnode()
    return hashed(str(mac_address))


def track(event_type: TrackingEventType, *args, **kwargs):
    """Should be the entry point to all tracking."""
    tracking_enabled = settings.get().TRACKING_ENABLED

    if not tracking_enabled:
        return

    switcher = {
        TrackingEventType.COPILOT_START: _track_copilot_start,
        TrackingEventType.CHAT_MESSAGE: _track_chat_message,
    }

    func = switcher.get(event_type)
    if func is None:
        # We cannot raise an exception (to not break the app), and probably warning would be too much spam here too? So currently failing silently.
        # Alternatively, we could say in warning that your tracking is broken and just turn it off with OPENCOPILOT_DO_NOT_TRACK
        return

    try:
        func(*args, **kwargs)
    except Exception as e:
        # Same reason to hide exception as above
        pass


def _track_copilot_start(
    n_documents: int,
    n_data_loaders: int,
    n_local_files_dirs: int,
    n_local_file_paths: int,
    n_data_urls: int,
):
    """Should be fired when the copilot starts."""
    s: Settings = settings.get()

    event = {
        "llm_name": s.MODEL,
        "copilot_name_hash": hashed(s.COPILOT_NAME),
        "prompt": {"hash": hashed(s.PROMPT), "length": len(s.PROMPT)},
        "retrieval": {
            "n_documents": n_documents,
            "n_data_loaders": n_data_loaders,
            "n_local_files_dirs": n_local_files_dirs,
            "n_data_urls": n_data_urls,
            "n_local_file_paths": n_local_file_paths,
        },
        "environment": {
            "python_version": platform.python_version(),
            "opencopilot_version": get_opencopilot_version(),
            "platform.platform": platform.platform(),
            "platform.system": platform.system(),
            # TODO track env type - conda, venv, docker, etc
        },
    }

    segment_analytics.track(
        anonymous_id=get_hashed_user_id(), event="Started Copilot", properties=event
    )


def _track_chat_message(user_agent, is_streaming):
    """Should be fired when a chat message is sent to the API."""
    copilot_name = settings.get().COPILOT_NAME
    event = {
        "is_streaming": is_streaming,
        "copilot_name_hash": hashed(copilot_name),
    }
    context = {
        "userAgent": user_agent,
    }

    segment_analytics.track(
        anonymous_id=get_hashed_user_id(),
        event="Messaged Copilot",
        properties=event,
        context=context,
    )



from typing import Optional

from fastapi import APIRouter
from fastapi import BackgroundTasks
from fastapi import Body
from fastapi import Depends
from fastapi import Request
from fastapi import Path
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
from pydantic import Field

from opencopilot.analytics import track
from opencopilot.analytics import TrackingEventType
from opencopilot.authorization import validate_api_key_use_case
from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.documents import document_store
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.routers import routing_utils
from opencopilot.service.chat import chat_conversations_service
from opencopilot.service.chat import chat_delete_service
from opencopilot.service.chat import chat_history_service
from opencopilot.service.chat import chat_service
from opencopilot.service.chat import chat_streaming_service
from opencopilot.service.chat.entities import ChatDeleteRequest
from opencopilot.service.chat.entities import ChatDeleteResponse
from opencopilot.service.chat.entities import ChatHistoryRequest
from opencopilot.service.chat.entities import ChatHistoryResponse
from opencopilot.service.chat.entities import ChatRequest
from opencopilot.service.chat.entities import ChatResponse
from opencopilot.service.chat.entities import ConversationsRequest
from opencopilot.service.chat.entities import ConversationsResponse

TAG = "Conversation"
router = APIRouter()
router.openapi_tags = [TAG]
router.title = "Conversation router"

logger = api_logger.get()

STREAM_RESPONSE_DESCRIPTION = """
A stream of objects, delimited by newlines. Each object will be of the following form:
```
{
    "text": "some text" # the next chunk of the message from the copilot
    "error": ""         # if present, a string description of the error that occurred
}
```

For example, the message "I like to eat apples" might be streamed as follows:

```
{"text": "I like"}
{"text": " to eat"}
{"text": " apples"}
```
"""

CONVERSATION_ID_DESCRIPTION = """
The ID of the conversation. To start a new conversation, you should pass in a random uuid version 4 (Python: `import uuid; uuid.uuid4()`). To continue a conversation, re-use the same uuid.
"""


class ConversationInput(BaseModel):
    message: str = Field(
        ...,
        description="Message to be answered by the copilot.",
        example="How do I make a delicious lemon cheesecake?",
    )
    response_message_id: Optional[str] = Field(
        None,
    )

    class Config:
        schema_extra = {
            "example": {
                "message": "How do I make a delicious lemon cheesecake?",
            }
        }


@router.get(
    "/conversations",
    summary="List conversations.",
    tags=[TAG],
    response_model=ConversationsResponse,
)
async def handle_get_conversations(
    user_id: str = Depends(validate_api_key_use_case.execute),
):
    users_repository = UsersRepositoryLocal()
    response = chat_conversations_service.execute(
        request=ConversationsRequest(user_id=user_id), users_repository=users_repository
    )
    return response


@router.post(
    "/conversations/{conversation_id}",
    summary="Send a message to the copilot and receive a non-streamed response.",
    tags=[TAG],
    response_model=ChatResponse,
)
async def handle_conversation(
    background_tasks: BackgroundTasks,
    api_request: Request,
    conversation_id: str = Path(
        ...,
        description=CONVERSATION_ID_DESCRIPTION,
    ),
    payload: ConversationInput = Body(
        ..., description="Input and parameters for the conversation."
    ),
    user_id: str = Depends(validate_api_key_use_case.execute),
):
    request = ChatRequest(
        conversation_id=conversation_id,
        message=payload.message,
        response_message_id=payload.response_message_id,
        user_id=user_id,
    )

    history_repository = ConversationHistoryRepositoryLocal()
    logs_repository = ConversationLogsRepositoryLocal()
    users_repository = UsersRepositoryLocal()

    response: ChatResponse = await chat_service.execute(
        request,
        document_store.get_document_store(),
        history_repository,
        logs_repository,
        users_repository,
    )

    background_tasks.add_task(
        track,
        TrackingEventType.CHAT_MESSAGE,
        api_request.headers.get("user-agent"),
        False,
    )
    return routing_utils.to_json_response(
        {"copilot_message": response.message, "sources": response.sources}
    )


@router.post(
    "/conversations/{conversation_id}/stream",
    summary="Send a message to the copilot and stream the response.",
    response_description=STREAM_RESPONSE_DESCRIPTION,
    tags=[TAG],
)
async def handle_conversation_streaming(
    background_tasks: BackgroundTasks,
    api_request: Request,
    conversation_id: str = Path(..., description=CONVERSATION_ID_DESCRIPTION),
    payload: ConversationInput = Body(
        ..., description="Input and parameters for the conversation."
    ),
    user_id: str = Depends(validate_api_key_use_case.execute),
):
    request = ChatRequest(
        conversation_id=conversation_id,
        message=payload.message,
        response_message_id=payload.response_message_id,
        user_id=user_id,
    )

    history_repository = ConversationHistoryRepositoryLocal()
    logs_repository = ConversationLogsRepositoryLocal()
    users_repository = UsersRepositoryLocal()

    headers = {
        "X-Content-Type-Options": "nosniff",
        "Connection": "keep-alive",
    }

    background_tasks.add_task(
        track,
        TrackingEventType.CHAT_MESSAGE,
        api_request.headers.get("user-agent"),
        True,
    )
    return StreamingResponse(
        chat_streaming_service.execute(
            request,
            document_store.get_document_store(),
            history_repository,
            logs_repository,
            users_repository,
        ),
        headers=headers,
        media_type="text/event-stream",
    )


@router.get(
    "/conversations/{conversation_id}",
    summary="Retrieve a conversation.",
    tags=[TAG],
    response_model=ChatHistoryResponse,
)
async def handle_get_conversation_history(
    conversation_id: str = Path(..., description="The ID of the conversation."),
    user_id: str = Depends(validate_api_key_use_case.execute),
):
    request = ChatHistoryRequest(
        conversation_id=conversation_id,
        user_id=user_id,
    )

    history_repository = ConversationHistoryRepositoryLocal()
    users_repository = UsersRepositoryLocal()

    response: ChatHistoryResponse = await chat_history_service.execute(
        request, history_repository, users_repository
    )
    return response


@router.delete(
    "/conversations/{conversation_id}",
    summary="Delete a conversation.",
    tags=[TAG],
    response_model=ChatDeleteResponse,
)
async def handle_delete_conversation_history(
    user_id: str = Depends(validate_api_key_use_case.execute),
    conversation_id: str = Path(..., description="The ID of the conversation."),
):
    users_repository = UsersRepositoryLocal()
    history_repository = ConversationHistoryRepositoryLocal()
    logs_repository = ConversationLogsRepositoryLocal()
    response = chat_delete_service.execute(
        request=ChatDeleteRequest(
            conversation_id=conversation_id,
            user_id=user_id,
        ),
        users_repository=users_repository,
        history_repository=history_repository,
        logs_repository=logs_repository,
    )
    return response



from dataclasses import dataclass
from typing import List

from langchain.chat_models.base import BaseChatModel


@dataclass(frozen=True)
class CacheObject:
    text: str
    token_count: int


cache: List[CacheObject] = []


def execute(
    text: str,
    llm: BaseChatModel,
    is_use_cache: bool = False,
) -> int:
    if not is_use_cache:
        return llm.get_num_tokens(text)

    for c in cache:
        if c.text == text:
            return c.token_count
    token_count = llm.get_num_tokens(text)
    cache.append(CacheObject(text=text, token_count=token_count))
    return token_count



import logging
import os
import uuid
from contextlib import nullcontext
from typing import Optional

import typer
from rich.console import Console
from rich.table import Table
from typing_extensions import Annotated

from opencopilot import exception_utils
from opencopilot.domain.cli import cli_chat_use_case
from opencopilot.logger import api_logger
from opencopilot.utils.scripting import set_default_settings

from opencopilot.oss import oss_app
from opencopilot.utils.validators import validate_openai_api_key

logger = api_logger.get()
logger.setLevel(logging.WARNING)
console = Console()

app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=False,
)

exception_utils.add_copilot_exception_catching()
app.add_typer(oss_app, name="oss")


@app.callback()
def main(ctx: typer.Context):
    # Initialize settings
    set_default_settings("cli")


@app.command(help="Chat with the Copilot. Example: chat 'Hello, who are you?'")
def chat(message: str):
    print("Message:", message)
    conversation_id = uuid.uuid4()
    while message:
        print("Response: ", end="", flush=True)
        cli_chat_use_case.conversation_stream(
            base_url="http://0.0.0.0:3000",
            conversation_id=conversation_id,
            message=message,
            stream=True,
        )
        print()
        message = input("Message: ")


@app.command(help="Query the retrieval pipeline.")
def retrieve(
    ctx: typer.Context,
    text: Annotated[
        Optional[str],
        typer.Option(
            "--text", "-t", help='Your question, i.e. "How to improve retrieval?"'
        ),
    ] = None,
    source: Annotated[
        Optional[str],
        typer.Option("--source", "-s", help="Source to match - supports wildcards"),
    ] = None,
    all: Annotated[
        Optional[bool], typer.Option("--all", "-a", help="Gets all documents ingested")
    ] = False,
):
    openai_api_key = os.getenv("OPENAI_API_KEY")
    validate_openai_api_key(openai_api_key)

    from opencopilot.repository.documents.document_store import WeaviateDocumentStore

    document_store = WeaviateDocumentStore()

    if text is not None:
        where_filter = (
            {"path": ["source"], "operator": "Like", "valueString": source}
            if source
            else None
        )
        document_chunks = document_store.find(text, where_filter=where_filter)
    elif source is not None:
        document_chunks = document_store.find_by_source(source)
    elif all:
        document_chunks = document_store.get_all()
    else:
        typer.echo(ctx.get_help())
        raise typer.Exit()

    documents = {}
    for chunk in document_chunks:
        source = chunk.metadata.get("source")
        if source:
            documents[source] = documents.get(source, 0) + 1

    table = Table("Source", "Chunks")
    for source in documents.keys():
        table.add_row(source, str(documents[source]))

    with nullcontext() if len(documents) < 50 else console.pager():
        console.print(table)
    print(f"{len(documents)} documents in {len(document_chunks)} chunks")


if __name__ == "__main__":
    app()



import json
import uuid

from opencopilot.domain.chat import on_user_message_streaming_use_case
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.documents.document_store import DocumentStore
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.chat.entities import ChatRequest
from opencopilot.service.utils import get_uuid


async def execute(
    request: ChatRequest,
    document_store: DocumentStore,
    history_repository: ConversationHistoryRepositoryLocal,
    logs_repository: ConversationLogsRepositoryLocal,
    users_repository: UsersRepositoryLocal,
) -> str:
    conversation_id = get_uuid(request.conversation_id, "conversation_id")
    async for chunk in on_user_message_streaming_use_case.execute(
        UserMessageInput(
            conversation_id=conversation_id,
            message=request.message,
            response_message_id=str(uuid.uuid4()),
            user_id=request.user_id,
        ),
        document_store,
        history_repository,
        logs_repository,
        users_repository,
    ):
        data = chunk.to_dict()
        yield f"{json.dumps(data)}\n"



import asyncio
import os
import webbrowser

from fastapi import FastAPI
from fastapi import Request
from fastapi.openapi.utils import get_openapi
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates
from pydantic import BaseModel
from starlette.middleware.cors import CORSMiddleware

import opencopilot
from opencopilot import settings
from opencopilot.routers import main_router
from opencopilot.routers import routing_utils
from opencopilot.service.exception_handlers.exception_handlers import (
    custom_exception_handler,
)
from opencopilot.service.middleware.main_middleware import MainMiddleware
from opencopilot.service.middleware.request_enrichment_middleware import (
    RequestEnrichmentMiddleware,
)

app = FastAPI()

app.include_router(main_router.router, prefix="/v0")

html_template_path = os.path.join(os.path.dirname(opencopilot.__file__), "html")
templates = Jinja2Templates(directory=html_template_path)

API_TITLE = "OpenCopilot API"
API_DESCRIPTION = (
    "OpenCopilot API, for more information visit https://docs.opencopilot.dev/"
)
API_VERSION = "0.1"

base_url = settings.get().get_base_url()


@app.on_event("startup")
async def startup_event():
    # TODO: once we support multiple workers move this to application.py
    loop = asyncio.get_event_loop()
    loop.create_task(_open_browser())


async def _open_browser():
    await asyncio.sleep(1)
    print(f"Started chat UI on {base_url}:{settings.get().API_PORT}/ui")
    try:
        cache_dir: str = os.path.expanduser("~/.opencopilot")
        if os.path.exists(cache_dir):
            return
        webbrowser.open(f"{base_url}:{settings.get().API_PORT}/ui", new=2)
        os.makedirs(cache_dir, exist_ok=True)
    except:
        pass


class ApiInfo(BaseModel):
    title: str
    description: str
    version: str

    class Config:
        schema_extra = {
            "example": {
                "title": API_TITLE,
                "description": API_DESCRIPTION,
                "version": API_VERSION,
            }
        }


def custom_openapi():
    if app.openapi_schema:
        return app.openapi_schema

    openapi_schema = get_openapi(
        title=API_TITLE,
        version=API_VERSION,
        description=API_DESCRIPTION,
        routes=app.routes,
        servers=_get_servers(),
    )
    openapi_schema["info"]["contact"] = {"name": "", "email": ""}
    openapi_schema["info"]["x-logo"] = {"url": ""}
    openapi_schema["x-readme"] = {
        "samples-languages": ["curl", "node", "javascript", "python"]
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


def _get_servers():
    servers = []
    if settings.get().is_production():
        pass
    else:
        servers.append({"url": f"{base_url}"})
    return servers


app.openapi = custom_openapi

# order of middleware matters! first middleware called is the last one added
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.get().ALLOWED_ORIGINS,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(MainMiddleware)
app.add_middleware(RequestEnrichmentMiddleware)

# exception handlers run AFTER the middlewares!
# Handles API error responses
app.add_exception_handler(Exception, custom_exception_handler)


# Overrides FastAPI error responses, eg: authorization, not found
# app.add_exception_handler(StarletteHTTPException, custom_http_exception_handler)
# Overrides default Pydantic request validation errors
# app.add_exception_handler(RequestValidationError, validation_exception_handler)


def get_api_info() -> ApiInfo:
    return ApiInfo(title=API_TITLE, description=API_DESCRIPTION, version=API_VERSION)


@app.get(
    "/",
    summary="Returns API information",
    description="Returns API information",
    response_description="API information with title, description and version.",
    response_model=ApiInfo,
    include_in_schema=not settings.get().is_production(),
)
def root():
    return routing_utils.to_json_response(get_api_info().dict())


@app.get("/ui", include_in_schema=False, response_class=HTMLResponse)
def ui(request: Request):
    return templates.TemplateResponse("index.html", {"request": request})



import asyncio
import json
from datetime import datetime
from typing import AsyncGenerator
from typing import List

from langchain.schema import Document

from opencopilot import settings
from opencopilot.domain.chat import is_user_allowed_to_chat_use_case
from opencopilot.domain.chat.entities import LoadingMessage
from opencopilot.domain.chat.entities import StreamingChunk
from opencopilot.domain.chat.entities import UserMessageInput
from opencopilot.domain.chat.results import get_gpt_result_use_case
from opencopilot.domain.chat.utils import get_system_message
from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.documents.document_store import DocumentStore
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.error_responses import ForbiddenAPIError
from opencopilot.utils.callbacks.callback_handler import (
    CustomAsyncIteratorCallbackHandler,
)

logger = api_logger.get()


async def execute(
    domain_input: UserMessageInput,
    document_store: DocumentStore,
    history_repository: ConversationHistoryRepositoryLocal,
    logs_repository: ConversationLogsRepositoryLocal,
    users_repository: UsersRepositoryLocal,
) -> AsyncGenerator[StreamingChunk, None]:
    if not is_user_allowed_to_chat_use_case.execute(
        domain_input.conversation_id,
        domain_input.user_id,
        history_repository,
        users_repository,
    ):
        raise ForbiddenAPIError()

    system_message = get_system_message()

    context = _get_context(domain_input, system_message, document_store)
    message_timestamp = datetime.now().timestamp()

    callback = CustomAsyncIteratorCallbackHandler()

    task = asyncio.create_task(
        get_gpt_result_use_case.execute(
            domain_input,
            system_message,
            context,
            logs_repository=logs_repository,
            history_repository=history_repository,
            callback=callback,
        )
    )

    result = ""
    try:
        async for callback_result in callback.aiter():
            parsed = json.loads(callback_result)
            if token := parsed.get("token"):
                yield StreamingChunk(
                    conversation_id=domain_input.conversation_id,
                    text=token,
                    sources=[],
                )
            if loading_message := parsed.get("loading_message"):
                yield StreamingChunk(
                    conversation_id=domain_input.conversation_id,
                    text="",
                    sources=[],
                    loading_message=LoadingMessage.from_dict(loading_message),
                )
        await task
        result = task.result()
    except Exception as exc:
        logger.error(f"Stream error: {exc}")
        yield StreamingChunk(
            conversation_id=domain_input.conversation_id,
            text="",
            sources=[],
            error=f"OpenAI error: {type(exc).__name__}",
        )
    finally:
        response_timestamp = datetime.now().timestamp()

        history_repository.save_history(
            domain_input.message,
            result,
            message_timestamp,
            response_timestamp,
            domain_input.conversation_id,
            domain_input.response_message_id,
        )
        users_repository.add_conversation(
            conversation_id=domain_input.conversation_id, user_id=domain_input.user_id
        )


def _get_context(
    domain_input: UserMessageInput, system_message: str, document_store: DocumentStore
) -> List[Document]:
    # TODO: handle context length and all the edge cases somehow a bit better
    context = []
    if "{context}" in system_message:
        context = []
        context.extend(
            document_store.find(
                domain_input.message,
                k=settings.get().MAX_CONTEXT_DOCUMENTS_COUNT - len(context),
            )
        )
    return context



import jwt
import time

from opencopilot import settings
from opencopilot.service import error_responses


def execute(client_id: str, client_secret: str, user_id: str) -> str:
    if (
        client_id != settings.get().JWT_CLIENT_ID
        or client_secret != settings.get().JWT_CLIENT_SECRET
    ):
        raise error_responses.InvalidCredentialsAPIError

    payload = {
        "iat": time.time() - 100,
        "exp": time.time() + settings.get().JWT_TOKEN_EXPIRATION_SECONDS,
        "sub": user_id,
    }
    return jwt.encode(payload, settings.get().JWT_CLIENT_SECRET, algorithm="HS256")



import json
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import requests
import aiohttp
from langchain.callbacks.manager import AsyncCallbackManagerForLLMRun
from langchain.callbacks.manager import CallbackManagerForLLMRun
from langchain.chat_models.base import BaseChatModel
from langchain.schema import BaseMessage
from langchain.schema import AIMessage
from langchain.schema import ChatResult
from langchain.schema import ChatGeneration
from pydantic import Extra

from urllib.parse import urljoin
from opencopilot.logger import api_logger

logger = api_logger.get()


class LocalLLM(BaseChatModel):
    context_size: int = 4096
    temperature: float = 0.7
    llm_url: str

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore
        allow_population_by_field_name = True

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        final = ""
        for text in self._get_stream(
            {"query": messages[0], "temperature": self.temperature, "stop": stop}
        ):
            final += self._process_text(text)
        return ChatResult(
            generations=[
                ChatGeneration(
                    message=AIMessage(content=final),
                )
            ]
        )

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        final = ""
        async for text in self._get_async_stream(
            {
                "query": messages[0].content,
                "temperature": self.temperature,
                "max_tokens": 0,
            }
        ):
            token = self._process_text(text)
            final += token
            if run_manager:
                await run_manager.on_llm_new_token(token)
        return ChatResult(
            generations=[
                ChatGeneration(
                    message=AIMessage(content=final),
                )
            ]
        )

    def get_token_ids(self, text: str) -> List[int]:
        try:
            result = requests.post(
                urljoin(self.llm_url, "/tokenize"), json={"text": text}
            )
            return result.json()["tokens"]
        except Exception as e:
            logger.error("Failed to get token count: %s", e)

    @property
    def _llm_type(self) -> str:
        return "local-llm"

    def _get_stream(self, payload: Dict):
        s = requests.Session()
        with s.post(
            urljoin(self.llm_url, "/generate_stream"), json=payload, stream=True
        ) as resp:
            for line in resp.iter_lines():
                if line:
                    yield line

    async def _get_async_stream(self, payload: Dict):
        timeout = aiohttp.ClientTimeout(sock_read=1200)
        async with aiohttp.ClientSession(timeout=timeout) as s:
            async with s.post(
                urljoin(self.llm_url, "/generate_stream"), json=payload
            ) as resp:
                async for line in resp.content.iter_any():
                    if line:
                        yield line

    def _process_text(self, line_raw):
        line = line_raw.decode("utf-8")
        try:
            line = line.replace('data:{"token"', '{"token"')
            line = json.loads(line)
            return line["token"]["text"]
        except:
            return ""



from abc import ABC


class CopilotConfigurationError(Exception, ABC):
    """Raised when the copilot configuration is invalid."""

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message


class PromptError(CopilotConfigurationError):
    """Raised when the prompt file passed in is missing or invalid."""


class APIKeyError(CopilotConfigurationError):
    """Raised when an API key is not provided, malformed, or rejected by the API provider."""

    pass


class ModelError(CopilotConfigurationError):
    """Raised when the model passed in is invalid."""

    pass


class LogsDirError(CopilotConfigurationError):
    """Raised when invalid logs dir was passed."""

    pass


class CopilotRuntimeError(Exception, ABC):
    """Raised when the copilot gets known run time exception."""

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message


class CopilotIsNotRunningError(CopilotRuntimeError):
    """Raised when cannot connect to Copilot"""


class OpenAIRuntimeError(CopilotRuntimeError):
    """Raised when cannot get a result from OpenAI"""


class WeaviateRuntimeError(CopilotRuntimeError):
    """Raised when cannot connect to Weaviate"""



from dataclasses import dataclass
from typing import Optional
from typing import Union

from langchain.chat_models.base import BaseChatModel
from langchain.embeddings.base import Embeddings

from opencopilot.domain.chat.models.local import LocalLLM


@dataclass(frozen=False)
class Settings:
    COPILOT_NAME: str

    HOST: str
    API_PORT: int
    ENVIRONMENT: str
    ALLOWED_ORIGINS: str

    LLM: Union[str, BaseChatModel]
    EMBEDDING_MODEL: Union[str, Embeddings]

    OPENAI_API_KEY: Optional[str] = None

    AUTH_TYPE: Optional[str] = None
    API_KEY: Optional[str] = None

    JWT_CLIENT_ID: Optional[str] = None
    JWT_CLIENT_SECRET: Optional[str] = None
    JWT_TOKEN_EXPIRATION_SECONDS: Optional[int] = 0

    HELICONE_API_KEY: Optional[str] = None
    HELICONE_RATE_LIMIT_POLICY: Optional[str] = None

    WEAVIATE_URL: Optional[str] = None
    WEAVIATE_READ_TIMEOUT: Optional[int] = 120

    MAX_DOCUMENT_SIZE_MB: Optional[int] = 50

    TRACKING_ENABLED: bool = False

    LOGS_DIR: str = "logs"
    # Configure based on model?
    PROMPT_HISTORY_INCLUDED_COUNT: int = 4
    MAX_CONTEXT_DOCUMENTS_COUNT: int = 4

    PROMPT: Optional[str] = None
    QUESTION_TEMPLATE: Optional[str] = None
    RESPONSE_TEMPLATE: Optional[str] = None

    HELICONE_BASE_URL = "https://oai.hconeai.com/v1"

    def __post_init__(self):
        if self.AUTH_TYPE is not None and (
            self.AUTH_TYPE == "none"
            or self.AUTH_TYPE == "None"
            or self.AUTH_TYPE.strip() == ""
        ):
            self.AUTH_TYPE = None

    def is_production(self):
        return self.ENVIRONMENT == "production"

    def get_max_token_count(self) -> int:
        if self.LLM == "gpt-3.5-turbo-16k":
            return 16384
        elif self.LLM == "gpt-4":
            return 8192
        elif isinstance(self.LLM, LocalLLM):
            return self.LLM.context_size
        return 2048

    def get_base_url(self) -> str:
        base_url = self.HOST
        if base_url.endswith("/"):
            base_url = base_url[:-1]
        base_url += ":" + str(self.API_PORT)
        if not base_url.startswith("http://"):
            base_url = "http://" + base_url
        return base_url


_settings: Optional[Settings] = None


def get() -> Optional[Settings]:
    global _settings
    return _settings


def set(new_settings: Settings):
    global _settings
    _settings = new_settings



from pydantic import BaseModel
from pydantic import Field


class ApiResponse(BaseModel):
    response: str = Field(description="Response status, either OK or NOK.")

    class Config:
        schema_extra = {"example": {"response": "OK"}}



from opencopilot.domain.chat import chat_conversations_use_case
from opencopilot.domain.chat.entities import ChatConversationsInput
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.chat.entities import ConversationsRequest
from opencopilot.service.chat.entities import ConversationsResponse


def execute(
    request: ConversationsRequest,
    users_repository: UsersRepositoryLocal,
) -> ConversationsResponse:
    result = chat_conversations_use_case.execute(
        data_input=ChatConversationsInput(
            user_id=request.user_id,
        ),
        users_repository=users_repository,
    )
    return ConversationsResponse(response="OK", conversation_ids=result.conversations)



import json
import os
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

import xxhash

from opencopilot import settings
from opencopilot.logger import api_logger

DEFAULT_USER_NAME = "default_user"

logger = api_logger.get()


class UsersRepositoryLocal:
    def __init__(self, users_dir: str = ""):
        if not users_dir:
            users_dir = os.path.join(settings.get().LOGS_DIR, "users")
        os.makedirs(users_dir, exist_ok=True)
        self.users_dir = users_dir

    def get_conversations(self, user_id: Optional[str] = None) -> List[str]:
        data = self._read_file(user_id)
        if data:
            return data.get("conversations") or []
        return []

    def add_conversation(
        self, conversation_id: UUID, user_id: Optional[str] = None
    ) -> None:
        data = self._read_file(user_id)
        if data.get("conversations"):
            conversations = set(data.get("conversations"))
            conversations.add(str(conversation_id))
            data["conversations"] = sorted(list(conversations))
        else:
            data["conversations"] = [str(conversation_id)]
        self._write_file(data, user_id)

    def remove_conversation(
        self, conversation_id: UUID, user_id: Optional[str] = None
    ) -> None:
        data = self._read_file(user_id)
        if data.get("conversations"):
            conversations = set(data.get("conversations"))
            if str(conversation_id) in conversations:
                conversations.remove(str(conversation_id))
                data["conversations"] = sorted(list(conversations))
                self._write_file(data, user_id)

    def _read_file(self, user_id: Optional[str] = None) -> Dict:
        try:
            file_path = self._get_file_path(user_id)
            with open(file_path, "r", encoding="utf-8") as file:
                return json.load(file)
        except FileNotFoundError:
            return {}

    def _write_file(self, data: Dict, user_id: Optional[str] = None):
        file_path = self._get_file_path(user_id)
        dir_path = os.path.dirname(file_path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path, exist_ok=True)
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(json.dumps(data, indent=4))

    def _get_file_path(self, user_id: Optional[str] = None) -> str:
        if not user_id:
            user_id = DEFAULT_USER_NAME
        file_name = xxhash.xxh64(user_id.encode("utf-8")).hexdigest()
        return os.path.join(self.users_dir, file_name) + ".json"



import os
from pathlib import Path
from typing import Optional

import requests
from dotenv import load_dotenv

HEADERS = {"accept": "application/json", "Content-Type": "application/json"}


def execute(base_url: str) -> Optional[str]:
    env_path = Path(".") / ".env"
    load_dotenv(dotenv_path=env_path)
    jwt_client_id = os.getenv("JWT_CLIENT_ID", "")
    jwt_client_secret = os.getenv("JWT_CLIENT_SECRET", "")
    url = f"{base_url}/v0/tokens"
    data = {
        "client_id": jwt_client_id,
        "client_secret": jwt_client_secret,
        "user_id": "test@local.host",
    }
    try:
        token_result = requests.post(url, headers=HEADERS, json=data)
        result_json = token_result.json()
        return result_json["token"]
    except:
        return None



from typing import Dict, Optional

import openai
from langchain.chat_models import ChatOpenAI
from langchain.chat_models.base import BaseChatModel

from opencopilot import settings


def execute(user_id: str = None, streaming: bool = False) -> BaseChatModel:
    llm = settings.get().LLM
    if isinstance(llm, str):
        if settings.get().HELICONE_API_KEY:
            openai.api_base = settings.get().HELICONE_BASE_URL
        return ChatOpenAI(
            temperature=0.0,
            model_name=settings.get().LLM,
            streaming=streaming,
            model_kwargs={"headers": _get_headers(user_id)},
            openai_api_key=settings.get().OPENAI_API_KEY,
        )
    if hasattr(llm, "streaming"):
        llm.streaming = streaming
    return llm


def _get_headers(user_id: str = None) -> Optional[Dict]:
    if settings.get().HELICONE_API_KEY:
        headers = {
            "Helicone-Auth": "Bearer " + settings.get().HELICONE_API_KEY,
            "Helicone-User-Id": user_id or "",
        }
        if user_id and settings.get().HELICONE_RATE_LIMIT_POLICY:
            headers[
                "Helicone-RateLimit-Policy"
            ] = settings.get().HELICONE_RATE_LIMIT_POLICY
        return headers
    return None



import time

from starlette.middleware.base import BaseHTTPMiddleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.types import ASGIApp

from opencopilot.domain.errors import CopilotRuntimeError
from opencopilot.domain.errors import OpenAIRuntimeError
from opencopilot.domain.errors import WeaviateRuntimeError
from opencopilot.logger import api_logger
from opencopilot.service.error_responses import APIErrorResponse
from opencopilot.service.error_responses import GenericCopilotRuntimeError
from opencopilot.service.error_responses import OpenAIError
from opencopilot.service.error_responses import WeaviateConnectionError
from opencopilot.service.middleware import util
from opencopilot.service.middleware.entities import RequestStateKey
from opencopilot.utils.http_headers import add_response_headers

logger = api_logger.get()


class MainMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        request_id = util.get_state(request, RequestStateKey.REQUEST_ID)
        ip_address = util.get_state(request, RequestStateKey.IP_ADDRESS)
        country = util.get_state(request, RequestStateKey.COUNTRY)
        try:
            logger.info(
                f"REQUEST STARTED "
                f"request_id={request_id} "
                f"request_path={request.url.path} "
                f"ip={ip_address} "
                f"country={country} "
            )
            before = time.time()
            response: Response = await call_next(request)

            duration = time.time() - before
            process_time = (time.time() - before) * 1000
            formatted_process_time = "{0:.2f}".format(process_time)
            if response.status_code != 404:
                logger.info(
                    f"REQUEST COMPLETED "
                    f"request_id={request_id} "
                    f"request_path={request.url.path} "
                    f"completed_in={formatted_process_time}ms "
                    f"status_code={response.status_code}"
                )
            return await _get_response_with_headers(response, duration)
        except OpenAIRuntimeError as exc:
            raise OpenAIError(exc.message)
        except WeaviateRuntimeError as exc:
            raise WeaviateConnectionError(exc.message)
        except CopilotRuntimeError as exc:
            raise GenericCopilotRuntimeError(exc.message)
        except Exception as error:
            if isinstance(error, APIErrorResponse):
                is_exc_info = error.to_status_code() == 500
                logger.error(
                    f"Error while handling request. request_id={request_id} "
                    f"request_path={request.url.path}"
                    f"status code={error.to_status_code()}"
                    f"code={error.to_code()}"
                    f"message={error.to_message()}",
                    exc_info=is_exc_info,
                )
            else:
                logger.error(
                    f"Error while handling request. request_id={request_id} "
                    f"request_path={request.url.path}",
                    exc_info=True,
                )
            raise error


async def _get_response_with_headers(response, duration):
    response = await add_response_headers(response, duration)
    return response



from typing import Dict
from typing import List

import trafilatura
from langchain.schema import Document


def execute(file_name: str, url: str) -> List[Document]:
    docs: List[Document] = []
    try:
        with open(file_name, "r") as f:
            content = f.read()
        text = trafilatura.extract(content)
        if not text:
            raise Exception("Failed to extract text")
        metadata: Dict = {
            "source": url,
        }
        extracted_metadata = trafilatura.extract_metadata(content)
        if extracted_metadata and extracted_metadata.title:
            metadata["title"] = extracted_metadata.title
        docs.append(Document(page_content=text, metadata=metadata))
    except Exception as e:
        raise e
    return docs



WEAVIATE_INVALID_URL = (
    "Invalid weaviate_url='{weaviate_url}' provided. Please make sure it starts with a "
    "schema like http:// or https://"
)

WEAVIATE_ERROR_EXTRA = (
    "\nPlease make sure that Weaviate is running and Copilot has data."
)

COPILOT_IS_NOT_RUNNING_ERROR = (
    "Could not connect to Copilot on url: '{copilot_url}"
    "\nPlease make sure that Copilot is running."
)

INVALID_MODEL_ERROR = (
    "Invalid llm_model_name='{llm_model_name}'.\n"
    "Allowed model names are: {allowed_model_names}"
)

INVALID_LOGS_DIR_ERROR = "Invalid logs_dir passed. Make sure it is non empty value."



from uuid import UUID
from opencopilot.service.error_responses import ValidationAPIError


def get_uuid(_id: str, param: str = "") -> UUID:
    try:
        return UUID(_id)
    except:
        raise ValidationAPIError(param=param)



"""
Search evaluation script.

Currently, the script expects that Weaviate vector store is running at http://0.0.0.0:8080/ and is
already filled with data.

If nothing is retrieved and nothing should be retrieved, then the result is perfect -
100% precision and 100% recall. But mostly, something should be retrieved. In that case:
If document retrieval returns nothing, the precision is 100%, and recall is 0%.
If document retrieval return everything, the precision is near 0%, and the recall is 100%.

Usage: see python eval_retrieval.py --help

"""
import argparse
import json
from typing import Callable
from typing import List


import matplotlib.pyplot as plt
from pathlib import Path
from langchain.schema import Document
from tqdm import tqdm

from opencopilot.utils.scripting import set_default_settings

set_default_settings("eval_retrieval")

from opencopilot.repository.documents import document_store
from opencopilot.eval.retrieval import evaluate_retrieval_dataset
from opencopilot.eval.entities import (
    RetrievalDataset,
    RetrievalResult,
    RetrievalSummaryEvaluation,
)


def parse_uri(s: str):
    """Turn into a (relatively) canonical URI."""
    s = s.strip()

    if "://" in s:
        return s[s.index("://") + 3 :]

    return s


def dataset_from_file(dataset_path: Path, limit: int = None) -> RetrievalDataset:
    with open(dataset_path) as f:
        examples_dict = json.load(f)
        if limit:
            examples_dict["examples"] = examples_dict["examples"][:limit]

        # Fix URLs - todo this is a bit hacky but best place to canonicalize URLs
        for example in examples_dict["examples"]:
            example["documents"] = [parse_uri(d) for d in example["documents"]]

    return RetrievalDataset.from_dict(examples_dict)


def print_metrics(metrics: RetrievalSummaryEvaluation):
    print("  Metrics")
    print(f"    average precision:", metrics.average_precision)
    print(f"       average recall:", metrics.average_recall)


def evaluate_retriever(
    search_function: Callable, dataset: RetrievalDataset, **search_kwargs
) -> RetrievalSummaryEvaluation:
    """Evaluate a retriever function on a dataset."""

    predictions = []
    for example in tqdm(dataset.examples):
        predicted_raw: List[Document] = search_function(example.query, **search_kwargs)
        retrieved_urls = [parse_uri(p.metadata["source"]) for p in predicted_raw]
        predictions.append(RetrievalResult(documents=retrieved_urls))

    return evaluate_retrieval_dataset(dataset, predictions)


def _draw_curve(dataset_path: str, limit=None):
    final_metrics = []
    for k in range(0, 11):
        document_retriever = document_store.get_document_store()
        search_type = f"similarity_search"
        metrics = evaluate_retriever(
            document_retriever.find,
            dataset=dataset_from_file(dataset_path, limit=limit),
            search_type=search_type,
            k=k + 1,
        )
        final_metrics.append(metrics)
        print_metrics(metrics)

    _plot_curve(final_metrics)


def _plot_curve(final_metrics: List[RetrievalSummaryEvaluation]):
    x = [item for item in range(len(final_metrics))]
    y1 = [item.average_precision for item in final_metrics]
    y2 = [item.average_recall for item in final_metrics]
    plt.plot(x, y1, label="average precision")
    plt.plot(x, y2, label="average recall")
    plt.xlabel("k")
    plt.ylabel("Average Precision & Recall")
    plt.title("Average Precision & Recall")
    plt.legend(loc="upper left")
    plt.show()


def _log_wandb(summary_evaluation: RetrievalSummaryEvaluation):
    import wandb

    wandb.init(
        # Set the project where this run will be logged
        project="opencopilot-retrieval",
        # Track hyperparameters and run metadata
        config={},
    )
    wandb.log(
        {
            "average_precision": summary_evaluation.average_precision,
            "average_recall": summary_evaluation.average_recall,
        }
    )
    with open(wandb.run.dir + "/evaluations.json", "w") as f:
        json.dump(summary_evaluation.to_dict(), f, indent=4)

    wandb.finish()


def main(
    dataset_path: str, draw_curve: bool, output_path: str = None, limit: int = None
):
    # TODO: document_store.init_document_store()
    if draw_curve:
        _draw_curve(dataset_path, limit=limit)
    else:
        document_retriever = document_store.get_document_store()
        metrics = evaluate_retriever(
            document_retriever.find,
            dataset=dataset_from_file(dataset_path, limit=limit),
            search_type="similarity_search",
            k=4,
        )
        print_metrics(metrics)
        if output_path:
            with open(output_path, "w") as f:
                json.dump(metrics.to_dict(), f, indent=4)
        return metrics


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset_path", type=str, help="Dataset path", required=True)
    parser.add_argument(
        "-n",
        "--num_examples",
        type=int,
        default=None,
        help="Limit how many examples are evaluated from dataset. Default: all examples used.",
    )
    parser.add_argument(
        "--draw_curve",
        action="store_true",
        help="Draw precision/recall curve. Disables --wandb and --output.",
    )
    parser.add_argument("--wandb", action="store_true", help="Store results in WANDB")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        default=None,
        help="Output file for evaluation results in JSON format.",
    )
    args = parser.parse_args()

    from opencopilot.repository.documents import document_store
    from opencopilot.repository.documents.document_store import WeaviateDocumentStore

    document_store.init_document_store(WeaviateDocumentStore())

    metrics = main(
        args.dataset_path,
        args.draw_curve,
        output_path=args.output,
        limit=args.num_examples,
    )

    if args.wandb and not args.draw_curve:
        print("Logging to WANDB:")
        _log_wandb(metrics)



from typing import List

from langchain.document_loaders import UnstructuredExcelLoader
from langchain.schema import Document


def execute(file_name: str, url: str) -> List[Document]:
    loader = UnstructuredExcelLoader(file_name)
    documents = loader.load()
    formatted_documents = []
    for document in documents:
        metadata = document.metadata or {}
        metadata["source"] = url
        formatted_documents.append(
            Document(page_content=document.page_content, metadata=metadata)
        )
    return formatted_documents



from typing import Optional
from uuid import UUID

from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.chat.entities import ChatHistoryItem

logger = api_logger.get()


async def execute(
    conversation_id: UUID,
    user_id: Optional[str],
    history_repository: ConversationHistoryRepositoryLocal,
    users_repository: UsersRepositoryLocal,
) -> [ChatHistoryItem]:
    if not await is_user_conversation(conversation_id, user_id, users_repository):
        return []

    response = history_repository.get_history(conversation_id)
    return_value = []
    timer = 0
    for message in response:
        prompt_timestamp = (
            message["prompt_timestamp"] if "prompt_timestamp" in message else timer
        )
        response_timestamp = (
            message["response_timestamp"]
            if "response_timestamp" in message
            else timer + 1
        )
        timer = max(timer, prompt_timestamp, response_timestamp) + 2
        return_value = return_value + [
            ChatHistoryItem(
                content=message["prompt"],
                timestamp=prompt_timestamp,
                response_message_id=message["response_message_id"],
            ),
            ChatHistoryItem(
                content=message["response"],
                timestamp=response_timestamp,
                response_message_id=message["response_message_id"],
            ),
        ]

    return return_value


async def is_user_conversation(
    conversation_id: UUID,
    user_id: Optional[str],
    users_repository: UsersRepositoryLocal,
) -> bool:
    conversations = users_repository.get_conversations(user_id)
    if str(conversation_id) in conversations:
        return True
    return False



import json
import os.path
from typing import Dict
from typing import List
from typing import Optional
from uuid import UUID

from opencopilot import settings
from opencopilot.logger import api_logger

logger = api_logger.get()


class ConversationHistoryRepositoryLocal:
    def __init__(
        self,
        conversations_dir: str = "",
        question_template: str = "",
        response_template: str = "",
    ):
        if not conversations_dir:
            conversations_dir = os.path.join(settings.get().LOGS_DIR, "conversations")
        self.conversations_dir = conversations_dir
        self.question_template = question_template or settings.get().QUESTION_TEMPLATE
        self.response_template = response_template or settings.get().RESPONSE_TEMPLATE

    def get_prompt_history(self, conversation_id: UUID, count: Optional[int]) -> str:
        try:
            with open(self._get_file_path(conversation_id), "r") as f:
                history = json.load(f)
            if not count or len(history) <= count:
                return self._to_string(history)
            return self._to_string(history[count * -1 :])
        except:
            logger.debug(
                f"Cannot load conversation history, id: {str(conversation_id)}"
            )
        return ""

    def get_history(self, conversation_id: UUID) -> List[Dict]:
        history = []
        try:
            with open(self._get_file_path(conversation_id), "r") as f:
                history = json.load(f)
        except:
            pass
        return history

    def _to_string(self, history: List[Dict]) -> str:
        formatted: str = ""
        for i in history:
            formatted += (
                f"{self.question_template.format(question=i.get('prompt', ''))}\n"
            )
            formatted += (
                f"{self.response_template.format(response=i.get('response', ''))}\n"
            )
        return formatted

    def save_history(
        self,
        message: str,
        result: str,
        prompt_timestamp: float,
        response_timestamp: float,
        conversation_id: UUID,
        response_message_id: str,
    ) -> None:
        history = self.get_history(conversation_id)
        history.append(
            {
                "prompt": message,
                "response": result.strip(),
                "prompt_timestamp": prompt_timestamp,
                "response_timestamp": response_timestamp,
                "response_message_id": response_message_id,
            }
        )
        self._write_file(conversation_id, history)

    def remove_conversation(self, conversation_id: UUID) -> None:
        file_path = self._get_file_path(conversation_id)
        if os.path.exists(file_path):
            try:
                os.remove(file_path)
            except:
                pass

    def _get_file_path(self, conversation_id: UUID):
        return f"{self.conversations_dir}/{str(conversation_id)}.json"

    def _write_file(self, conversation_id: UUID, data):
        try:
            if not os.path.exists(self.conversations_dir):
                os.makedirs(self.conversations_dir, exist_ok=True)
            with open(self._get_file_path(conversation_id), "w") as f:
                f.write(json.dumps(data, indent=4))
        except Exception:
            logger.warning(f"Failed to save history for chat {str(conversation_id)}")



from opencopilot.domain.chat.entities import ChatConversationsInput
from opencopilot.domain.chat.entities import ChatConversationsOutput
from opencopilot.repository.users_repository import UsersRepositoryLocal


def execute(
    data_input: ChatConversationsInput, users_repository: UsersRepositoryLocal
) -> ChatConversationsOutput:
    result = users_repository.get_conversations(data_input.user_id)
    return ChatConversationsOutput(conversations=result)



import os
import platform

from setuptools import setup


def package_files(directory):
    paths = []
    for path, directories, filenames in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


def set_env():
    if platform.system() == "Darwin":
        os.system('export CMAKE_ARGS="-DLLAMA_METAL=on"')


set_env()
extra_files = package_files("opencopilot")

setup(
    name='opencopilot-ai',
    version='0.3.6',
    packages=["opencopilot"],
    package_data={"": extra_files},
    license="MIT",
    description="OpenCopilot Backend",
    author="OpenCopilot",
    author_email="kaspar@nftport.xyz",
    url="https://github.com/opencopilotdev/opencopilot",
    py_modules=["opencopilot"],
    install_requires=[
        "fastapi",
        "dataclasses-json",
        "python-dotenv",
        "python-json-logger",
        "uvicorn",
        "pandas",
        "pexpect",
        "langchain>=0.0.217",
        "passlib",
        "pyjwt[crypto]",
        "Jinja2",
        "tiktoken",
        "weaviate-client>=3.23.2",
        "pypdf",
        "unstructured",
        "pdf2image",
        "openai",
        "unstructured",
        "xxhash",
        "omegaconf",
        "openpyxl",
        "typer[all]",
        "trafilatura",
        "filetype",
        "segment-analytics-python",
        "psutil",
    ],
    extras_require={
        "llama": [
            "llama-cpp-python==0.1.83",
        ],
    },
    entry_points={
        "console_scripts": [
            "opencopilot = opencopilot.cli:app",
        ],
    },
)



from opencopilot.domain.chat.entities import ChatDeleteInput
from opencopilot.domain.chat.entities import ChatDeleteOutput
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal


def execute(
    data_input: ChatDeleteInput,
    users_repository: UsersRepositoryLocal,
    history_repository: ConversationHistoryRepositoryLocal,
    logs_repository: ConversationLogsRepositoryLocal,
) -> ChatDeleteOutput:
    conversations = users_repository.get_conversations(data_input.user_id)
    if str(data_input.conversation_id) in conversations:
        history_repository.remove_conversation(data_input.conversation_id)
        logs_repository.remove_conversation(data_input.conversation_id)
        users_repository.remove_conversation(
            conversation_id=data_input.conversation_id, user_id=data_input.user_id
        )
        return ChatDeleteOutput(response="OK")
    return ChatDeleteOutput(response="NOK")



import os
import platform
from dataclasses import dataclass
from typing import Any
from typing import Optional

import typer
import psutil
import requests
from tqdm import tqdm
from rich.console import Console
from rich.table import Table
from rich import print

console = Console()

oss_app = typer.Typer(
    name="oss",
    help="OpenCopilot tool to manage and interact with Open Source LLMs.",
    no_args_is_help=True,
)

MODEL_PATH = os.path.expanduser("~/.opencopilot/models/")

LLAMA_PROMPT_TEMPLATE = """<s>[INST] <<SYS>>\nYour purpose is to repeat what the user says, but in a different wording.\nDon't add anything, don't answer any questions, don't give any advice or comment - just repeat.\nContext:\n{context}\n<</SYS>>\n\n{history} Repeat: {question} [/INST]"""

CODELLAMA_PROMPT_TEMPLATE = "<s>[INST] <<SYS>>\nWrite code to solve the following coding problem that obeys the constraints and passes the example test cases.\nPlease wrap your code answer using ```.\nRelevant information: {context}. \n<</SYS>>\n\n{history} {question} [/INST]"


@dataclass
class ModelInfo:
    name: str
    size: float
    description: str
    prompt_template: str
    filename: str
    url: str
    context_size: int


MODELS = {
    "llama-2-7b-chat": ModelInfo(
        name="llama-2-7b-chat",
        size=3.83,
        description="Meta developed and publicly released the Llama 2 family of large language models (LLMs), a collection of pretrained and fine-tuned generative text models ranging in scale from 7 billion to 70 billion parameters. Fine-tuned LLMs, called Llama-2-Chat, are optimized for dialogue use cases.",
        prompt_template=LLAMA_PROMPT_TEMPLATE,
        filename="llama-2-7b-chat.Q4_0.gguf",
        context_size=4096,
        url="https://huggingface.co/TheBloke/Llama-2-7b-Chat-GGUF/resolve/main/llama-2-7b-chat.Q4_0.gguf",
    ),
    "llama-2-13b-chat": ModelInfo(
        name="Llama-2-13b-chat",
        size=7.37,
        description="Meta developed and publicly released the Llama 2 family of large language models (LLMs), a collection of pretrained and fine-tuned generative text models ranging in scale from 7 billion to 70 billion parameters. Fine-tuned LLMs, called Llama-2-Chat, are optimized for dialogue use cases.",
        prompt_template=LLAMA_PROMPT_TEMPLATE,
        filename="llama-2-13b-chat.Q4_0.gguf",
        context_size=4096,
        url="https://huggingface.co/TheBloke/Llama-2-13B-chat-GGUF/resolve/main/llama-2-13b-chat.Q4_0.gguf",
    ),
    "llama-2-70b-chat": ModelInfo(
        name="Llama-2-70b-chat",
        size=38.9,
        description="Meta developed and publicly released the Llama 2 family of large language models (LLMs), a collection of pretrained and fine-tuned generative text models ranging in scale from 7 billion to 70 billion parameters. Fine-tuned LLMs, called Llama-2-Chat, are optimized for dialogue use cases.",
        prompt_template=LLAMA_PROMPT_TEMPLATE,
        filename="llama-2-70b-chat.Q4_0.gguf",
        context_size=4096,
        url="https://huggingface.co/TheBloke/Llama-2-70B-chat-GGUF/resolve/main/llama-2-70b-chat.Q4_0.gguf",
    ),
    "codellama-7b": ModelInfo(
        name="CodeLlama-7b",
        size=3.83,
        description="Code Llama is a collection of pretrained and fine-tuned generative text models ranging in scale from 7 billion to 34 billion parameters. This model is designed for general code synthesis and understanding.",
        prompt_template=CODELLAMA_PROMPT_TEMPLATE,
        filename="codellama-7b.Q4_0.gguf",
        context_size=4096,
        url="https://huggingface.co/TheBloke/CodeLlama-7B-GGUF/resolve/main/codellama-7b.Q4_0.gguf",
    ),
    "codellama-13b": ModelInfo(
        name="CodeLlama-13b",
        size=7.37,
        description="Code Llama is a collection of pretrained and fine-tuned generative text models ranging in scale from 7 billion to 34 billion parameters. This model is designed for general code synthesis and understanding.",
        prompt_template=CODELLAMA_PROMPT_TEMPLATE,
        filename="codellama-13b.Q4_0.gguf",
        context_size=4096,
        url="https://huggingface.co/TheBloke/CodeLlama-13B-GGUF/resolve/main/codellama-13b.Q4_0.gguf",
    ),
    "codellama-34b": ModelInfo(
        name="CodeLlama-34b",
        size=19.1,
        description="Code Llama is a collection of pretrained and fine-tuned generative text models ranging in scale from 7 billion to 34 billion parameters. This model is designed for general code synthesis and understanding.",
        prompt_template=CODELLAMA_PROMPT_TEMPLATE,
        filename="codellama-34b.Q4_0.gguf",
        context_size=4096,
        url="https://huggingface.co/TheBloke/CodeLlama-34B-GGUF/resolve/main/codellama-34b.Q4_0.gguf",
    ),
}


def _is_macos() -> bool:
    return platform.system() == "Darwin"


def _can_use_model(model: ModelInfo) -> bool:
    total_memory = psutil.virtual_memory().total / (1024**3)
    return model.size < total_memory / 2


def _is_model_installed(model: ModelInfo) -> bool:
    return os.path.exists(os.path.join(MODEL_PATH, model.filename))


def _remove_model(model: ModelInfo) -> bool:
    return os.remove(os.path.join(MODEL_PATH, model.filename))


def _print_model_not_found_message(model_name: str):
    print(f"[red]Model [code]{model_name}[/code] is not available.[/red]")
    names = [f"[code]{m}[/code]" for m in MODELS]
    available_models = ", ".join(names[:-1]) + " and " + names[-1]
    print(f"Available models: {available_models}")


def _download_model(url: str, filename: str):
    model_file_path = os.path.join(MODEL_PATH, filename)
    if os.path.exists(model_file_path):
        resume_byte = os.path.getsize(model_file_path)
    else:
        resume_byte = 0

    headers = {}
    if resume_byte:
        headers["Range"] = f"bytes={resume_byte}-"

    response = requests.get(url, headers=headers, stream=True)

    total_size = resume_byte + int(response.headers.get("content-length", 0))
    progress_bar = tqdm(
        total=total_size, unit="B", unit_scale=True, initial=resume_byte
    )

    os.makedirs(MODEL_PATH, exist_ok=True)
    with open(model_file_path, "ab" if resume_byte else "wb") as f:
        for chunk in response.iter_content(chunk_size=8192):
            f.write(chunk)
            progress_bar.update(len(chunk))
    progress_bar.close()


def _try_llama_cpp_import() -> Any:
    # pylint: disable=import-error
    try:
        import llama_cpp

        return llama_cpp
    except:
        print(
            "Cannot run LLM; make sure you've installed the [code]llama-cpp-python[/code] package and dependencies!"
        )
        if _is_macos():
            print(
                'To install: [code]CMAKE_ARGS="-DLLAMA_METAL=on" pip install llama-cpp-python[/code]'
            )
        else:
            print("To install: [code]pip install llama-cpp-python[/code]")
        print(
            "More information on how to install: [link]https://llama-cpp-python.readthedocs.io/en/latest/#installation[/link]"
        )
        print("Re-run this command after installation is done!")
        return None


@oss_app.command("list")
def list_models():
    """List available open source large language models"""
    table = Table(
        "",
        "NAME",
        "SIZE",
        "INSTALLED",
    )
    for model_name, model in MODELS.items():
        table.add_row(
            "*" if _can_use_model(model) else "",
            model_name,
            f"{model.size}GB",
            "Yes" if _is_model_installed(model) else "No",
        )
    console.print(table)
    print("\n* Recommended for your system")
    print(
        "\nTo see more details about a model: [code]opencopilot oss info <model_name>[/code]"
    )


@oss_app.command("info")
def model_info(model_name: str):
    try:
        model = MODELS.get(model_name.lower())
        table = Table(show_header=False, box=None)
        table.add_column("Label", no_wrap=True, style="bold")
        table.add_column("Value")
        table.add_row("Model Name:", model.name)
        table.add_row("Size:", f"{model.size} GB")
        table.add_row("Description:", model.description)
        console.print(table)
    except:
        _print_model_not_found_message(model_name)


@oss_app.command("remove")
def model_remove(model_name: str):
    model = MODELS.get(model_name.lower())
    if not model:
        _print_model_not_found_message(model_name)
        return
    if _is_model_installed(model):
        _remove_model(model)
        print(f"LLM [bold]{model.name}[/bold] removed successfully.")
    else:
        print(f"[bold]{model.name}[/bold] not downloaded - nothing to remove.")


@oss_app.command("run")
def run_model(
    model_name: Optional[str] = typer.Argument("Llama-2-7b-chat"),
    host: Optional[str] = typer.Option(
        None, "--host", help="Hostname to run the LLM on."
    ),
    port: Optional[int] = typer.Option(None, "--port", help="Port to run the LLM on."),
):
    """Run a specific model."""
    llama_cpp = _try_llama_cpp_import()
    if not llama_cpp:
        return

    model = MODELS.get(model_name.lower())
    if not model:
        _print_model_not_found_message(model_name)
        return

    try:
        print(
            f"Preparing to download the {model_name} model from Huggingface. This might be a large file."
        )
        print(f"The model will be saved to: {MODEL_PATH}.")
        print(
            f"We appreciate your patience. If the download gets interrupted, don't worry, you can always resume it later."
        )
        _download_model(model.url, model.filename)
        print(f"Now loading {model.name}. Hang tight!")
    except KeyboardInterrupt:
        print("[red]Download interrupted.[/red]")
        print(
            f"You can resume it by running [code]opencopilot oss run {model_name}[/code] again."
        )
        return
    except:
        print(f"[red]Could not run {model_name}![/red]")
        return

    import uvicorn
    from opencopilot.oss_llm.app import create_app

    app = create_app(
        model=os.path.join(MODEL_PATH, model.filename),
        context_size=model.context_size,
    )
    uvicorn.run(
        app,
        host=host if host else "localhost",
        port=port if port else 8000,
    )


@oss_app.command("prompt")
def generate_prompt(model_name: str):
    """Generate a model-specific prompt template."""
    if model_name.lower() in MODELS:
        typer.echo(MODELS[model_name].prompt_template)
    else:
        _print_model_not_found_message(model_name)



from typing import Optional
from uuid import UUID

from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal


def execute(
    conversation_id: UUID,
    user_id: Optional[str],
    history_repository: ConversationHistoryRepositoryLocal,
    users_repository: UsersRepositoryLocal,
) -> bool:
    conversation = history_repository.get_history(conversation_id)
    if not conversation:
        return True
    elif _is_user_conversation(conversation_id, user_id, users_repository):
        return True
    return False


def _is_user_conversation(
    conversation_id: UUID,
    user_id: Optional[str],
    users_repository: UsersRepositoryLocal,
) -> bool:
    conversations = users_repository.get_conversations(user_id)
    if str(conversation_id) in conversations:
        return True
    return False



from pydantic import BaseModel
from pydantic import Field
from typing import List


class GenerateStreamRequest(BaseModel):
    query: str
    temperature: float = 0.1
    max_tokens: int


class TokenizeRequest(BaseModel):
    text: str


class TokenizeResponse(BaseModel):
    tokens: List[int] = Field(description="List of tokens")

    class Config:
        schema_extra = {"example": {"tokens": [1, 52, 332, 44, 16]}}



from fastapi import APIRouter

from opencopilot.logger import api_logger
from opencopilot.routers.routes import chat_router
from opencopilot.routers.routes import debug_router, auth_router

TAG_ROOT = "root"

router = APIRouter()
logger = api_logger.get()

routers_to_include = [
    chat_router.router,
    debug_router.router,
    auth_router.router,
]
route_titles = []
for router_to_include in routers_to_include:
    router.include_router(router_to_include)
    route_titles.append(
        {"title": router_to_include.title, "tags": router_to_include.openapi_tags}
    )



from fastapi import APIRouter
from fastapi import Depends
from fastapi import Path

from opencopilot.authorization import validate_api_key_use_case
from opencopilot.logger import api_logger
from opencopilot.repository.conversation_history_repository import (
    ConversationHistoryRepositoryLocal,
)
from opencopilot.repository.conversation_logs_repository import (
    ConversationLogsRepositoryLocal,
)
from opencopilot.repository.users_repository import UsersRepositoryLocal
from opencopilot.service.debug import message_debug_service
from opencopilot.service.debug.entities import GetMessageDebugResponse

TAG = "Conversation"
router = APIRouter()
router.openapi_tags = [TAG]
router.title = "Debug router"

logger = api_logger.get()


@router.get(
    "/debug/{conversation_id}/{message_id}",
    tags=[TAG],
    summary="Get debug information about a message.",
    response_model=GetMessageDebugResponse,
)
async def get_copilots(
    conversation_id: str = Path(..., description="The ID of the conversation."),
    message_id: str = Path(..., description="The ID of the response message."),
    user_id: str = Depends(validate_api_key_use_case.execute),
):
    history_repository = ConversationHistoryRepositoryLocal()
    logs_repository = ConversationLogsRepositoryLocal()
    users_repository = UsersRepositoryLocal()

    return message_debug_service.execute(
        conversation_id,
        message_id,
        history_repository,
        logs_repository,
        users_repository=users_repository,
        user_id=user_id,
    )



from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from opencopilot.service.entities import ApiResponse


class ConversationsRequest(BaseModel):
    user_id: Optional[str] = None


class ConversationsResponse(ApiResponse):
    conversation_ids: List[str]

    class Config:
        schema_extra = {
            "example": {
                "response": "OK",
                "conversation_ids": ["e91042aa-d53a-41eb-8884-67aa4947982d"],
            }
        }


class ChatRequest(BaseModel):
    conversation_id: str = Field(description="Conversation id")
    message: str
    user_id: Optional[str] = None


class ChatResponse(ApiResponse):
    conversation_id: str = Field(description="Conversation id")

    message: str = Field(description="Conversation output")

    sources: List[str] = Field(default_factory=list, description="Sources")

    class Config:
        schema_extra = {
            "example": {
                "response": "OK",
                "conversation_id": "e91042aa-d53a-41eb-8884-67aa4947982d",
                "message": "I will use the 'search' command to find the weather in San Francisco.",
            }
        }


class ChatHistoryRequest(BaseModel):
    conversation_id: str = Field(description="Chat id")
    user_id: Optional[str] = None


class ChatHistoryItem(BaseModel):
    content: str
    timestamp: float
    response_message_id: str


class ChatHistoryResponse(BaseModel):
    response: str
    conversation_id: str = Field(description="Chat id")
    messages: List[ChatHistoryItem] = Field(
        default_factory=list, description="Messages"
    )

    class Config:
        schema_extra = {
            "example": {
                "response": "OK",
                "conversation_id": "e91042aa-d53a-41eb-8884-67aa4947982d",
                "messages": [
                    {
                        "content": "Hello",
                        "timestamp": 1693562530,
                        "response_message_id": "ed02eedf-7a74-4a31-8fbf-eeb4300faf31",
                    },
                    {
                        "content": "Hello, how are you?",
                        "timestamp": 1693562539,
                        "response_message_id": "ed02eedf-7a74-4a31-8fbf-eeb4300faf31",
                    },
                ],
            }
        }


class ChatDeleteRequest(BaseModel):
    conversation_id: str = Field(description="Chat id")
    user_id: Optional[str] = None


class ChatDeleteResponse(ApiResponse):
    pass
